# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the engine.

Sub-Modules:

.. currentmodule:: co2mpas.core.model.physical.engine

.. autosummary::
    :nosignatures:
    :toctree: engine/

    fc
    idle
    thermal
"""
import math
import functools
import numpy as np
import schedula as sh
from co2mpas.defaults import dfl
from .idle import dsp as _idle
from .thermal import dsp as _thermal
from .fc import dsp as _fc

dsp = sh.BlueDispatcher(name='Engine', description='Models the vehicle engine.')


@sh.add_function(dsp, outputs=['fuel_type', 'is_hybrid'])
def define_fuel_type_and_is_hybrid(obd_fuel_type_code):
    """
    Defines engine fuel type and if the vehicle is hybrid.

    :param obd_fuel_type_code:
        OBD fuel type of the vehicle [-].
    :type obd_fuel_type_code: int

    :return:
        Engine fuel type and if the vehicle is hybrid.
    :rtype: (str, bool)
    """

    i = int(obd_fuel_type_code)
    d = dfl.functions.define_fuel_type_and_is_hybrid
    return d.fuel_type.get(i, sh.NONE), d.is_hybrid.get(i, sh.NONE)


@sh.add_function(dsp, outputs=['engine_mass'])
def calculate_engine_mass(ignition_type, engine_max_power):
    """
    Calculates the engine mass [kg].

    :param ignition_type:
        Engine ignition type (positive or compression).
    :type ignition_type: str

    :param engine_max_power:
        Engine nominal power [kW].
    :type engine_max_power: float

    :return:
       Engine mass [kg].
    :rtype: float
    """

    par = dfl.functions.calculate_engine_mass.PARAMS
    _mass_coeff = par['mass_coeff']
    m, q = par['mass_reg_coeff']
    # Engine mass empirical formula based on web data found for engines weighted
    # according DIN 70020-GZ
    # kg
    return (m * engine_max_power + q) * _mass_coeff[ignition_type]


@sh.add_function(dsp, outputs=['engine_heat_capacity'])
def calculate_engine_heat_capacity(engine_mass):
    """
    Calculates the engine heat capacity [kg*J/K].

    :param engine_mass:
        Engine mass [kg].
    :type engine_mass: float

    :return:
       Engine heat capacity [kg*J/K].
    :rtype: float
    """

    par = dfl.functions.calculate_engine_heat_capacity.PARAMS
    mp, hc = par['heated_mass_percentage'], par['heat_capacity']

    return engine_mass * np.sum([hc[k] * v for k, v in mp.items()])


@sh.add_function(dsp, outputs=['ignition_type'])
def default_ignition_type(engine_type):
    """
    Returns the default ignition type according to the fuel type.

    :param engine_type:
        Engine type (positive turbo, positive natural aspiration, compression).
    :type engine_type: str

    :return:
        Engine ignition type (positive or compression).
    :rtype: str
    """

    if 'compression' in engine_type:
        return 'compression'
    return 'positive'


@sh.add_function(dsp, outputs=['ignition_type'], weight=1)
def default_ignition_type_v1(fuel_type):
    """
    Returns the default ignition type according to the fuel type.

    :param fuel_type:
        Fuel type (diesel, gasoline, LPG, NG, ethanol, methanol, biodiesel,
        propane).
    :type fuel_type: str

    :return:
        Engine ignition type (positive or compression).
    :rtype: str
    """

    if 'diesel' in fuel_type:
        return 'compression'
    return 'positive'


@sh.add_function(dsp, outputs=['full_bmep_curve'])
def define_full_bmep_curve(
        full_load_speeds, full_load_curve, min_engine_on_speed,
        engine_capacity, engine_stroke, idle_engine_speed, engine_max_speed):
    """
    Defines the vehicle full bmep curve.

    :param full_load_speeds:
        T1 map speed vector [RPM].
    :type full_load_speeds: numpy.array

    :param full_load_curve:
        Vehicle full load curve.
    :type full_load_curve: function

    :param min_engine_on_speed:
        Minimum engine speed to consider the engine to be on [RPM].
    :type min_engine_on_speed: float

    :param engine_capacity:
        Engine capacity [cm3].
    :type engine_capacity: float

    :param engine_stroke:
        Engine stroke [mm].
    :type engine_stroke: float

    :param idle_engine_speed:
        Idle engine speed and its standard deviation [RPM].
    :type idle_engine_speed: (float, float)

    :param engine_max_speed:
        Maximum allowed engine speed [RPM].
    :type engine_max_speed: float

    :return:
        Vehicle full bmep curve.
    :rtype: function
    """

    speeds = np.unique(np.append(
        full_load_speeds, [idle_engine_speed[0], engine_max_speed]
    ))
    from .fc import calculate_brake_mean_effective_pressures
    p = calculate_brake_mean_effective_pressures(
        speeds, full_load_curve(speeds), engine_capacity,
        min_engine_on_speed)

    s = calculate_mean_piston_speeds(speeds, engine_stroke)

    return functools.partial(np.interp, xp=s, fp=p, left=0, right=0)


dsp.add_data('is_cycle_hot', dfl.values.is_cycle_hot)


@sh.add_function(dsp, outputs=['full_load_powers'])
def calculate_full_load_powers(full_load_torques, full_load_speeds):
    """
    Calculates T1 map power vector [kW].

    :param full_load_torques:
        T1 map torque vector [N*m].
    :type full_load_torques: numpy.array

    :param full_load_speeds:
        T1 map speed vector [RPM].
    :type full_load_speeds: numpy.array

    :return:
        T1 map power vector [kW].
    :rtype: numpy.array
    """
    from ..wheels import calculate_wheel_powers
    return calculate_wheel_powers(full_load_torques, full_load_speeds)


@sh.add_function(dsp, outputs=['full_load_speeds'])
def calculate_full_load_speeds(full_load_powers, full_load_torques):
    """
    Calculates T1 map speed vector [RPM].

    :param full_load_powers:
        T1 map power vector [kW].
    :type full_load_powers: numpy.array

    :param full_load_torques:
        T1 map torque vector [N*m].
    :type full_load_torques: numpy.array

    :return:
        T1 map speed vector [RPM].
    :rtype: numpy.array
    """
    from ..wheels import calculate_wheel_torques
    return calculate_wheel_torques(full_load_powers, full_load_torques)


@sh.add_function(dsp, outputs=['engine_max_speed'], weight=20)
def default_engine_max_speed(
        ignition_type, idle_engine_speed, engine_speed_at_max_power):
    """
    Returns the default maximum allowed engine speed [RPM].

    :param ignition_type:
        Engine ignition type (positive or compression).
    :type ignition_type: str

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :param engine_speed_at_max_power:
        Engine speed at engine nominal power [RPM].
    :type engine_speed_at_max_power: float

    :return:
        Maximum allowed engine speed [RPM].
    :rtype: float
    """
    fl = dfl.functions.default_full_load_speeds_and_powers.FULL_LOAD
    idl, r = idle_engine_speed[0], max(fl[ignition_type][0])
    return idl + r * (engine_speed_at_max_power - idl)


@sh.add_function(dsp, outputs=['full_load_speeds', 'full_load_powers'])
def default_full_load_speeds_and_powers(
        ignition_type, engine_max_power, engine_speed_at_max_power,
        idle_engine_speed, engine_max_speed):
    """
    Returns the defaults full load speeds and powers [RPM, kW].

    :param ignition_type:
        Engine ignition type (positive or compression).
    :type ignition_type: str

    :param engine_max_power:
        Engine nominal power [kW].
    :type engine_max_power: float

    :param engine_speed_at_max_power:
        Engine speed at engine nominal power [RPM].
    :type engine_speed_at_max_power: float

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :param engine_max_speed:
        Maximum allowed engine speed [RPM].
    :type engine_max_speed: float

    :return:
         T1 map speed [RPM] and power [kW] vectors.
    :rtype: (numpy.array, numpy.array)
    """
    from scipy.interpolate import InterpolatedUnivariateSpline as Spline

    idl = idle_engine_speed[0]

    full_load_speeds = np.unique(np.append(
        [engine_speed_at_max_power], np.linspace(idl, engine_max_speed)
    ))

    d = dfl.functions.default_full_load_speeds_and_powers
    full_load_powers = Spline(*d.FULL_LOAD[ignition_type], k=1)(
        (full_load_speeds - idl) / (engine_speed_at_max_power - idl)
    ) * engine_max_power

    return full_load_speeds, full_load_powers


@sh.add_function(dsp, outputs=['engine_speed_at_max_power'])
def identify_engine_speed_at_max_power(full_load_speeds, full_load_powers):
    """
    Identifies engine nominal speed at engine nominal power [RPM].

    :param full_load_speeds:
        T1 map speed vector [RPM].
    :type full_load_speeds: numpy.array

    :param full_load_powers:
        T1 map power vector [kW].
    :type full_load_powers: numpy.array

    :return:
        Engine speed at engine nominal power [RPM].
    :rtype: float
    """
    return full_load_speeds[np.argmax(full_load_powers)]


@sh.add_function(dsp, outputs=['engine_max_power'])
def calculate_engine_max_power(full_load_curve, engine_speed_at_max_power):
    """
    Calculates engine nominal power [kW].

    :param full_load_curve:
        Vehicle full load curve.
    :type full_load_curve: function

    :param engine_speed_at_max_power:
        Engine speed at engine nominal power [RPM].
    :type engine_speed_at_max_power: float

    :return:
        Engine nominal power [kW].
    :rtype: float
    """
    return full_load_curve(engine_speed_at_max_power)


@sh.add_function(dsp, outputs=['full_load_curve'])
def define_full_load_curve(
        full_load_speeds, full_load_powers, idle_engine_speed,
        engine_max_speed):
    """
    Calculates the full load curve.

    :param full_load_speeds:
        T1 map speed vector [RPM].
    :type full_load_speeds: numpy.array

    :param full_load_powers:
        T1 map power vector [kW].
    :type full_load_powers: numpy.array

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :param engine_max_speed:
        Maximum allowed engine speed [RPM].
    :type engine_max_speed: float

    :return:
        Vehicle full load curve.
    :rtype: function
    """
    xp = [idle_engine_speed[0] - idle_engine_speed[1], engine_max_speed]
    xp.extend(full_load_speeds)
    xp = np.unique(xp)
    fp = np.interp(xp, full_load_speeds, full_load_powers)
    return functools.partial(np.interp, xp=xp, fp=fp, left=0, right=0)


@sh.add_function(dsp, outputs=['engine_max_speed'])
def identify_engine_max_speed(full_load_speeds):
    """
    Identifies the maximum allowed engine speed [RPM].

    :param full_load_speeds:
        T1 map speed vector [RPM].
    :type full_load_speeds: numpy.array

    :return:
        Maximum allowed engine speed [RPM].
    :rtype: float
    """
    return np.max(full_load_speeds)


@sh.add_function(dsp, outputs=['engine_max_torque'])
def calculate_engine_max_torque(
        engine_max_power, engine_speed_at_max_power, ignition_type):
    """
    Calculates engine nominal torque [N*m].

    :param engine_max_power:
        Engine nominal power [kW].
    :type engine_max_power: float

    :param engine_speed_at_max_power:
        Engine speed at engine nominal power [RPM].
    :type engine_speed_at_max_power: float

    :param ignition_type:
        Engine ignition type (positive or compression).
    :type ignition_type: str

    :return:
        Engine nominal torque [N*m].
    :rtype: float
    """

    c = dfl.functions.calculate_engine_max_torque.PARAMS[ignition_type]
    pi = math.pi
    return engine_max_power / engine_speed_at_max_power * (30000.0 / pi * c)


@sh.add_function(dsp, outputs=['engine_max_power'])
def calculate_engine_max_power_v1(
        engine_max_torque, engine_speed_at_max_power, ignition_type):
    """
    Calculates engine nominal power [kW].

    :param engine_max_torque:
        Engine nominal torque [N*m].
    :type engine_max_torque: float

    :param engine_speed_at_max_power:
        Engine speed at engine nominal power [RPM].
    :type engine_speed_at_max_power: float

    :param ignition_type:
        Engine ignition type (positive or compression).
    :type ignition_type: str

    :return:
        Engine nominal power [kW].
    :rtype: float
    """

    c = calculate_engine_max_torque(1, engine_speed_at_max_power, ignition_type)

    return engine_max_torque / c


dsp.add_dispatcher(
    include_defaults=True,
    dsp=_idle,
    inputs=(
        'idle_engine_speed_median', 'min_engine_on_speed', 'engine_speeds_out',
        'idle_engine_speed_std', 'idle_engine_speed', 'velocities',
        'stop_velocity'
    ),
    outputs=(
        'idle_engine_speed_median', 'idle_engine_speed_std',
        'idle_engine_speed'
    )
)

dsp.add_dispatcher(
    include_defaults=True,
    dsp=_thermal,
    dsp_id='thermal',
    inputs=(
        'max_engine_coolant_temperature', 'initial_engine_temperature', 'times',
        'engine_thermostat_temperature_window', 'engine_thermostat_temperature',
        'engine_temperature_regression_model', 'idle_engine_speed', 'on_engine',
        'engine_coolant_temperatures', 'gear_box_powers_out', 'accelerations',
        'engine_speeds_out', 'velocities', 'after_treatment_warm_up_phases',
        'temperature_shift'
    ),
    outputs=(
        'engine_thermostat_temperature_window', 'engine_thermostat_temperature',
        'engine_temperature_regression_model', 'max_engine_coolant_temperature',
        'engine_temperature_derivatives', 'engine_coolant_temperatures',
        'initial_engine_temperature', 'engine_temperatures', 'temperature_shift'
    )
)


@sh.add_function(
    dsp,
    inputs=[
        'on_engine', 'idle_engine_speed', 'engine_speeds_out_hot',
        'after_treatment_speeds_delta', 'clutch_tc_speeds_delta'
    ],
    outputs=['engine_speeds_out']
)
def calculate_engine_speeds_out(
        on_engine, idle_engine_speed, engine_speeds_out_hot, *delta_speeds):
    """
    Calculates the engine speed [RPM].

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :param idle_engine_speed:
        Idle engine speed and its standard deviation [RPM].
    :type idle_engine_speed: (float, float)

    :param engine_speeds_out_hot:
        Engine speed at hot condition [RPM].
    :type engine_speeds_out_hot: numpy.array

    :return:
        Engine speed [RPM].
    :rtype: numpy.array
    """

    speeds = engine_speeds_out_hot.copy()
    s = speeds[on_engine]
    for delta in delta_speeds:
        s += delta[on_engine]

    dn = idle_engine_speed[0]

    s[s < dn] = dn

    speeds[on_engine] = s

    return speeds


@sh.add_function(dsp, outputs=['engine_inertia_powers_losses'])
def calculate_engine_inertia_powers_losses(
        times, engine_speeds_out, engine_moment_inertia):
    """
    Calculates the engine power losses due to inertia [kW].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param engine_moment_inertia:
        Engine moment of inertia [kg*m2].
    :type engine_moment_inertia: float

    :param engine_speeds_out:
        Engine speed [RPM].
    :type engine_speeds_out: numpy.array

    :return:
        Engine power losses due to inertia [kW].
    :rtype: numpy.array
    """
    t = times[:, None] + np.array([-1, 1])
    c = engine_moment_inertia / 2000 * (2 * math.pi / 60) ** 2 / 4
    return c * np.diff(np.interp(t, times, engine_speeds_out)).ravel() ** 2


dsp.add_data('belt_efficiency', dfl.values.belt_efficiency)
dsp.add_function(
    function=sh.bypass,
    inputs=['belt_efficiency'],
    outputs=['belt_mean_efficiency']
)


@sh.add_function(dsp, outputs=['gross_engine_powers_out'])
def calculate_gross_engine_powers_out(
        clutch_tc_powers, belt_efficiency, on_engine, gear_box_type, on_idle,
        alternator_powers, motor_p0_powers, motor_p1_powers,
        motor_p2_planetary_powers):
    """
    Calculates the gross engine power (pre-losses) [kW].

    :param clutch_tc_powers:
        Clutch or torque converter power [kW].
    :type clutch_tc_powers: numpy.array

    :param belt_efficiency:
        Belt efficiency [-].
    :type belt_efficiency: float

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :param gear_box_type:
        Gear box type (manual or automatic or cvt).
    :type gear_box_type: str

    :param on_idle:
        If the engine is on idle [-].
    :type on_idle: numpy.array

    :param alternator_powers:
        Alternator power [kW].
    :type alternator_powers: numpy.array

    :param motor_p0_powers:
        Power at motor P0 [kW].
    :type motor_p0_powers: numpy.array

    :param motor_p1_powers:
        Power at motor P1 [kW].
    :type motor_p1_powers: numpy.array

    :param motor_p2_planetary_powers:
        Power at planetary motor P2 [kW].
    :type motor_p2_planetary_powers: numpy.array

    :return:
        Gross engine power (pre-losses) [kW].
    :rtype: numpy.array
    """

    p, b = np.zeros_like(clutch_tc_powers, dtype=float), on_engine
    p[b] = clutch_tc_powers[b]

    if gear_box_type == 'manual':
        p[on_idle & (p < 0)] = 0.0
    eff = np.where(motor_p0_powers[b] < 0, belt_efficiency, 1 / belt_efficiency)
    p[b] -= alternator_powers[b] + motor_p0_powers[b] * eff
    p[b] -= motor_p1_powers[b] + motor_p2_planetary_powers[b]
    return p


@sh.add_function(dsp, outputs=['clutch_tc_powers'])
def calculate_clutch_tc_powers(
        gross_engine_powers_out, belt_efficiency, alternator_powers,
        motor_p0_powers, motor_p1_powers, motor_p2_planetary_powers):
    """
    Calculates the clutch or torque converter power [kW] from gross engine kW.

    :param gross_engine_powers_out:
        Gross engine power (pre-losses) [kW].
    :type gross_engine_powers_out: numpy.array

    :param belt_efficiency:
        Belt efficiency [-].
    :type belt_efficiency: float

    :param alternator_powers:
        Alternator power [kW].
    :type alternator_powers: numpy.array

    :param motor_p0_powers:
        Power at motor P0 [kW].
    :type motor_p0_powers: numpy.array

    :param motor_p1_powers:
        Power at motor P1 [kW].
    :type motor_p1_powers: numpy.array

    :param motor_p2_planetary_powers:
        Power at planetary motor P2 [kW].
    :type motor_p2_planetary_powers: numpy.array

    :return:
        Clutch or torque converter power [kW].
    :rtype: numpy.array
    """
    p = np.where(motor_p0_powers < 0, belt_efficiency, 1 / belt_efficiency)
    p *= motor_p0_powers
    p += gross_engine_powers_out + alternator_powers + motor_p1_powers
    p += motor_p2_planetary_powers
    return p


@sh.add_function(dsp, outputs=['min_available_engine_powers_out'])
def calculate_min_available_engine_powers_out(
        engine_stroke, engine_capacity, initial_friction_params,
        engine_speeds_out):
    """
    Calculates the minimum available engine power (i.e., engine motoring curve).

    :param engine_stroke:
        Engine stroke [mm].
    :type engine_stroke: float

    :param engine_capacity:
        Engine capacity [cm3].
    :type engine_capacity: float

    :param initial_friction_params:
        Engine initial friction params l & l2 [-].
    :type initial_friction_params: float, float

    :param engine_speeds_out:
        Engine speed [RPM].
    :type engine_speeds_out: numpy.array | float

    :return:
        Minimum available engine power [kW].
    :rtype: numpy.array | float
    """

    l, l2 = np.array(initial_friction_params) * (engine_capacity / 1200000.0)
    l2 *= (engine_stroke / 30000.0) ** 2

    return (l2 * engine_speeds_out * engine_speeds_out + l) * engine_speeds_out


@sh.add_function(dsp, outputs=['max_available_engine_powers_out'])
def calculate_max_available_engine_powers_out(
        full_load_curve, engine_speeds_out):
    """
    Calculates the maximum available engine power [kW].

    :param full_load_curve:
        Vehicle full load curve.
    :type full_load_curve: function

    :param engine_speeds_out:
        Engine speed vector [RPM].
    :type engine_speeds_out: numpy.array | float

    :return:
        Maximum available engine power [kW].
    :rtype: numpy.array | float
    """

    return full_load_curve(engine_speeds_out)


@sh.add_function(dsp, outputs=['gross_engine_powers_out'])
def calculate_gross_engine_powers_out_v1(
        engine_powers_out, auxiliaries_power_losses,
        engine_inertia_powers_losses):
    """
    Calculates the gross engine power (pre-losses) [kW].

    :param engine_powers_out:
        Engine power vector [kW].
    :type engine_powers_out: numpy.array

    :param engine_inertia_powers_losses:
        Engine power losses due to inertia [kW].
    :type engine_inertia_powers_losses: numpy.array

    :param auxiliaries_power_losses:
        Engine power losses due to engine auxiliaries [kW].
    :type auxiliaries_power_losses: numpy.array

    :return:
        Gross engine power (pre-losses) [kW].
    :rtype: numpy.array
    """
    gross_engine_powers_out = engine_powers_out - auxiliaries_power_losses
    gross_engine_powers_out -= engine_inertia_powers_losses
    return gross_engine_powers_out


@sh.add_function(
    dsp, outputs=['engine_powers_out', 'missing_powers', 'brake_powers']
)
def correct_engine_powers_out(
        max_available_engine_powers_out, min_available_engine_powers_out,
        gross_engine_powers_out, auxiliaries_power_losses,
        engine_inertia_powers_losses):
    """
    Corrects the engine powers out according to the available powers and
    returns the missing and brake power [kW].

    :param max_available_engine_powers_out:
        Maximum available engine power [kW].
    :type max_available_engine_powers_out: numpy.array

    :param min_available_engine_powers_out:
        Minimum available engine power [kW].
    :type min_available_engine_powers_out: numpy.array

    :param engine_inertia_powers_losses:
        Engine power losses due to inertia [kW].
    :type engine_inertia_powers_losses: numpy.array

    :param auxiliaries_power_losses:
        Engine power losses due to engine auxiliaries [kW].
    :type auxiliaries_power_losses: numpy.array

    :param gross_engine_powers_out:
        Gross engine power (pre-losses) [kW].
    :type gross_engine_powers_out: numpy.array

    :return:
        Engine, missing, and braking powers [kW].
    :rtype: numpy.array, numpy.array, numpy.array
    """

    ul, dl = max_available_engine_powers_out, min_available_engine_powers_out
    p = gross_engine_powers_out + auxiliaries_power_losses
    p += engine_inertia_powers_losses
    up, dn = ul < p, dl > p

    missing_powers, brake_powers = np.zeros_like(p), np.zeros_like(p)
    missing_powers[up], brake_powers[dn] = p[up] - ul[up], dl[dn] - p[dn]

    return np.where(up, ul, np.where(dn, dl, p)), missing_powers, brake_powers


@sh.add_function(dsp, outputs=['has_sufficient_power'])
def check_vehicle_has_sufficient_power(times, missing_powers):
    """
    Checks if the vehicle has sufficient power.

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param missing_powers:
        Missing powers [kW].
    :type missing_powers: numpy.array

    :return:
        The cycle's percentage in which the vehicle has sufficient power [%].
    :rtype: float
    """
    w = np.zeros_like(times, dtype=float)
    t = (times[:-1] + times[1:]) / 2
    # noinspection PyUnresolvedReferences
    w[0], w[1:-1], w[-1] = t[0] - times[0], np.diff(t), times[-1] - t[-1]
    return 1 - np.average(missing_powers != 0, weights=w)


@sh.add_function(dsp, outputs=['mean_piston_speeds'])
def calculate_mean_piston_speeds(engine_speeds_out, engine_stroke):
    """
    Calculates mean piston speed [m/sec].

    :param engine_speeds_out:
        Engine speed vector [RPM].
    :type engine_speeds_out: numpy.array

    :param engine_stroke:
        Engine stroke [mm].
    :type engine_stroke: float

    :return:
        Mean piston speed vector [m/s].
    :rtype: numpy.array | float
    """

    return (engine_stroke / 30000.0) * engine_speeds_out


dsp.add_data('engine_is_turbo', dfl.values.engine_is_turbo)


@sh.add_function(dsp, outputs=['engine_type'])
def calculate_engine_type(ignition_type, engine_is_turbo):
    """
    Calculates the engine type (gasoline turbo, gasoline natural aspiration,
    diesel).

    :param ignition_type:
        Engine ignition type (positive or compression).
    :type ignition_type: str

    :param engine_is_turbo:
        If the engine is equipped with any kind of charging.
    :type engine_is_turbo: bool

    :return:
        Engine type (positive turbo, positive natural aspiration, compression).
    :rtype: str
    """

    engine_type = ignition_type

    if ignition_type == 'positive':
        engine_type = 'turbo' if engine_is_turbo else 'natural aspiration'
        engine_type = '%s %s' % (ignition_type, engine_type)

    return engine_type


@sh.add_function(dsp, outputs=['engine_moment_inertia'])
def calculate_engine_moment_inertia(engine_capacity, ignition_type):
    """
    Calculates engine moment of inertia [kg*m2].

    :param engine_capacity:
        Engine capacity [cm3].
    :type engine_capacity: float

    :param ignition_type:
        Engine ignition type (positive or compression).
    :type ignition_type: str

    :return:
        Engine moment of inertia [kg*m2].
    :rtype: float
    """
    par = dfl.functions.calculate_engine_moment_inertia.PARAMS[ignition_type]
    return (0.05 + 0.1 * engine_capacity / 1000.0) * par


dsp.add_data(
    'auxiliaries_torque_loss_factors',
    dfl.values.auxiliaries_torque_loss_factors
)
dsp.add_data('auxiliaries_power_loss', dfl.values.auxiliaries_power_loss)


@sh.add_function(dsp, outputs=['auxiliaries_torque_loss'])
def calculate_auxiliaries_torque_loss(
        auxiliaries_torque_loss_factors, engine_capacity):
    """
    Calculates engine torque losses due to engine auxiliaries [N*m].

    :param auxiliaries_torque_loss_factors:
        Constant torque loss factors due to engine auxiliaries [N/m2, N*m].
    :type auxiliaries_torque_loss_factors: (float, float)

    :param engine_capacity:
        Engine capacity [cm3].
    :type engine_capacity: float

    :return:
        Engine torque losses due to engine auxiliaries [N*m].
    :rtype: numpy.array
    """
    m, q = auxiliaries_torque_loss_factors
    return m * engine_capacity / 1000.0 + q


@sh.add_function(dsp, outputs=['auxiliaries_torque_losses'])
def calculate_auxiliaries_torque_losses(times, auxiliaries_torque_loss):
    """
    Calculates engine torque losses due to engine auxiliaries [N*m].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param auxiliaries_torque_loss:
        Constant torque loss due to engine auxiliaries [N*m].
    :type auxiliaries_torque_loss: float

    :return:
        Engine torque losses due to engine auxiliaries [N*m].
    :rtype: numpy.array
    """

    return np.tile(auxiliaries_torque_loss, len(times))


@sh.add_function(dsp, outputs=['auxiliaries_power_losses'])
def calculate_auxiliaries_power_losses(
        auxiliaries_torque_losses, engine_speeds_out, on_engine,
        auxiliaries_power_loss):
    """
    Calculates engine power losses due to engine auxiliaries [kW].

    :param auxiliaries_torque_losses:
        Engine torque losses due to engine auxiliaries [N*m].
    :type auxiliaries_torque_losses: numpy.array

    :param engine_speeds_out:
        Engine speed [RPM].
    :type engine_speeds_out: numpy.array

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :param auxiliaries_power_loss:
        Constant power loss due to engine auxiliaries [kW].
    :type auxiliaries_power_loss: float

    :return:
        Engine power losses due to engine auxiliaries [kW].
    :rtype: numpy.array
    """

    from ..wheels import calculate_wheel_powers
    p = calculate_wheel_powers(auxiliaries_torque_losses, engine_speeds_out)
    if auxiliaries_power_loss:
        p[on_engine] += auxiliaries_power_loss
    return p


dsp.add_dispatcher(
    include_defaults=True,
    dsp=_fc,
    dsp_id='fc_model',
    inputs=(
        'engine_has_variable_valve_actuation', 'engine_max_speed', 'velocities',
        'engine_thermostat_temperature', 'engine_thermostat_temperature_window',
        'engine_n_cylinders', 'min_engine_on_speed', 'phases_integration_times',
        'has_selective_catalytic_reduction', 'engine_has_cylinder_deactivation',
        'engine_fuel_lower_heating_value', 'idle_engine_speed', 'co2_emissions',
        'calibration_status', 'fuel_consumptions', 'engine_powers_out', 'times',
        'engine_speeds_out', 'full_load_powers', 'full_bmep_curve', 'on_engine',
        'full_load_speeds', 'engine_capacity', 'stop_velocity', 'has_lean_burn',
        'has_exhausted_gas_recirculation', 'fuel_carbon_content', 'engine_type',
        'engine_temperatures', 'mean_piston_speeds', 'phases_distances',
        'co2_normalization_references', 'phases_co2_emissions', 'is_cycle_hot',
        'active_cylinder_ratios', 'engine_stroke', 'co2_params', 'fuel_type',
        'after_treatment_warm_up_phases', 'engine_inertia_powers_losses',
        'fuel_consumptions_liters', 'is_hybrid', 'fuel_density', {
            'co2_params_calibrated': ('co2_params_calibrated', 'co2_params'),
            'engine_idle_fuel_consumption': (
                'engine_idle_fuel_consumption',
                'idle_fuel_consumption_initial_guess'
            )
        }),
    outputs=(
        'active_lean_burns', 'co2_rescaling_scores', 'co2_params_initial_guess',
        'engine_idle_fuel_consumption',
        'identified_co2_emissions', 'co2_emissions_model', 'calibration_status',
        'initial_friction_params', 'active_variable_valves', 'active_cylinders',
        'fuel_carbon_content', 'fuel_consumptions', 'co2_emissions', 'fuel_map',
        'extended_phases_integration_times', 'mean_piston_speeds', 'fmep_model',
        'active_exhausted_gas_recirculations', 'extended_phases_co2_emissions',
        'has_exhausted_gas_recirculation', 'engine_fuel_lower_heating_value',
        'co2_params_calibrated', 'fuel_consumptions_liters'
    ),
    inp_weight={'co2_params': dfl.EPS}
)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and constants to define the engine_coolant_temperature_model selector.
"""
import schedula as sh
import co2mpas.utils as co2_utl
from ._core import define_sub_model
from ...physical.engine.thermal import dsp as _thermal

#: Model name.
name = 'engine_coolant_temperature_model'

#: Parameters that constitute the model.
models = [
    'engine_temperature_regression_model', 'max_engine_coolant_temperature',
    'engine_thermostat_temperature', 'temperature_shift'
]

#: Inputs required to run the model.
inputs = [
    'times', 'on_engine', 'velocities', 'engine_speeds_out',
    'accelerations', 'initial_engine_temperature',
    'after_treatment_warm_up_phases'
]

#: Relevant outputs of the model.
outputs = ['engine_coolant_temperatures']

#: Targets to compare the outputs of the model.
targets = outputs

#: Metrics to compare outputs with targets.
metrics = sh.map_list(targets, co2_utl.mae)

#: Upper score limits to raise the warnings.
up_limit = sh.map_list(targets, 4)

#: Prediction model.
# noinspection PyProtectedMember
dsp = sh.Blueprint(_thermal, inputs, outputs, models)._set_cls(define_sub_model)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the A/T gear shifting.

Sub-Modules:

.. currentmodule:: co2mpas.core.model.physical.gear_box.at_gear

.. autosummary::
    :nosignatures:
    :toctree: at_gear/

    cmv
    cmv_ch
    core
    dtgs
    gspv
    gspv_ch
"""

import copy
import itertools
import collections
import numpy as np
import schedula as sh
from co2mpas.defaults import dfl
import co2mpas.utils as co2_utl
from .core import define_gear_filter
from .cmv import dsp as _cmv, CMV
from .cmv_ch import dsp as _cmv_ch
from .gspv import dsp as _gspv
from .gspv_ch import dsp as _gspv_ch
from .dtgs import dsp as _dtgs

dsp = sh.BlueDispatcher(
    name='Automatic gear model',
    description='Defines an omni-comprehensive gear shifting model for '
                'automatic vehicles.'
)
dsp.add_data(
    'fuel_saving_at_strategy', dfl.values.fuel_saving_at_strategy,
    description='Apply the eco-mode gear shifting?'
)
dsp.add_data('plateau_acceleration', dfl.values.plateau_acceleration)


# noinspection PyMissingOrEmptyDocstring
class MVL(CMV):
    def __init__(self, *args,
                 plateau_acceleration=dfl.values.plateau_acceleration,
                 **kwargs):
        super(MVL, self).__init__(*args, **kwargs)
        self.plateau_acceleration = plateau_acceleration

    # noinspection PyMethodOverriding,PyMethodOverriding
    def fit(self, gears, velocities, velocity_speed_ratios, idle_engine_speed,
            stop_velocity):
        self.velocity_speed_ratios = velocity_speed_ratios
        idle = idle_engine_speed
        mvl = [np.array([idle[0] - idle[1], idle[0] + idle[1]])]
        for k in range(1, int(max(gears)) + 1):
            lm, on, vsr = [], None, velocity_speed_ratios[k]

            for i, b in enumerate(itertools.chain(gears == k, [False])):
                if not b and on is not None:
                    v = velocities[on:i]
                    lm.append([min(v), max(v)])
                    on = None

                elif on is None and b:
                    on = i

            if lm:
                min_v, max_v = zip(*lm)
                lm = [sum(co2_utl.reject_outliers(min_v)), max(max_v)]
                mvl.append(np.array([max(idle[0], l / vsr) for l in lm]))
            else:
                mvl.append(mvl[-1].copy())

        mvl = [[k, tuple(v * velocity_speed_ratios[k])]
               for k, v in reversed(list(enumerate(mvl[1:], 1)))]
        mvl[0][1] = (mvl[0][1][0], dfl.INF)
        mvl.append([0, (0, mvl[-1][1][0])])

        for i, v in enumerate(mvl[1:]):
            v[1] = (v[1][0], max(v[1][1], mvl[i][1][0] + stop_velocity))

        self.clear()
        # noinspection PyTypeChecker
        self.update(collections.OrderedDict(mvl))

        return self

    # noinspection PyMethodOverriding
    def predict(self, velocity, acceleration, gear):
        if abs(acceleration) < self.plateau_acceleration:
            for k, v in self.items():
                if k <= gear:
                    break
                elif velocity > v[0]:
                    return k

        if gear:
            while velocity > self[gear][1]:
                gear += 1

        return gear


@sh.add_function(dsp, outputs=['MVL'])
def calibrate_mvl(
        gears, velocities, velocity_speed_ratios, idle_engine_speed,
        stop_velocity):
    """
    Calibrates the matrix velocity limits (upper and lower bound) [km/h].

    :param gears:
        Gear vector [-].
    :type gears: numpy.array

    :param velocities:
        Vehicle velocity [km/h].
    :type velocities: numpy.array

    :param velocity_speed_ratios:
        Constant velocity speed ratios of the gear box [km/(h*RPM)].
    :type velocity_speed_ratios: dict[int | float]

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :param stop_velocity:
        Maximum velocity to consider the vehicle stopped [km/h].
    :type stop_velocity: float

    :return:
        Matrix velocity limits (upper and lower bound) [km/h].
    :rtype: MVL
    """

    mvl = MVL().fit(gears, velocities, velocity_speed_ratios, idle_engine_speed,
                    stop_velocity)

    return mvl


dsp.add_func(
    define_gear_filter, inputs_kwargs=True, inputs_defaults=True,
    outputs=['gear_filter']
)


# noinspection PyMissingOrEmptyDocstring,PyAttributeOutsideInit,PyUnusedLocal
class CorrectGear:
    def __init__(self, velocity_speed_ratios=None, idle_engine_speed=None):
        velocity_speed_ratios = velocity_speed_ratios or {}
        self.gears = np.array(sorted(k for k in velocity_speed_ratios if k > 0))
        self.vsr = velocity_speed_ratios
        self.min_gear = velocity_speed_ratios and self.gears[0] or None
        self.idle_engine_speed = idle_engine_speed
        self.pipe = []

    def fit_basic_correct_gear(self):
        idle = self.idle_engine_speed[0] - self.idle_engine_speed[1]
        self.idle_vel = [(self.vsr[k] * idle, k) for k in self.gears[::-1]]
        self.pipe.append(self.basic_correct_gear)

    def basic_correct_gear(
            self, gear, i, gears, times, velocities, accelerations,
            motive_powers, engine_temperatures, next_gear):

        v, vel = velocities[i], velocities

        if v < self.idle_vel[-1][0]:
            gear = 0
        else:
            for iv, g in self.idle_vel:
                if gear >= g and v >= iv:
                    gear = g
                    break

        if not gear and v > 0:
            if accelerations[i] > 0 or (len(vel) > i + 1 and vel[i + 1] > v):
                gear = self.min_gear
        return gear

    def fit_correct_gear_mvl(self, mvl):
        self.mvl = mvl
        self.pipe.append(self.correct_gear_mvl)

    # noinspection PyUnusedLocal
    def correct_gear_mvl(
            self, gear, i, gears, times, velocities, accelerations,
            motive_powers, engine_temperatures, next_gear):
        return self.mvl.predict(velocities[i], accelerations[i], gear)

    def fit_correct_gear_full_load(
            self, full_load_curve, max_velocity_full_load_correction):
        self.max_velocity_full_load_corr = max_velocity_full_load_correction
        self.flc = full_load_curve
        self.np_vsr = np.array(list(map(self.vsr.get, self.gears)))
        self.pipe.append(self.correct_gear_full_load)

    def correct_gear_full_load(
            self, gear, i, gears, times, velocities, accelerations,
            motive_powers, engine_temperatures, next_gear):
        if motive_powers is None:
            return gear
        vel = velocities[i]
        if vel > self.max_velocity_full_load_corr or gear <= self.min_gear:
            return gear

        j = np.searchsorted(self.gears, gear)
        delta = self.flc(vel / self.np_vsr) - motive_powers[i]
        valid = delta[:j + 1][::-1] >= 0
        k = valid.argmax()
        if not valid[k]:
            return self.gears[np.argmax(delta)]
        return self.gears[j - k]

    def fit_correct_driveability_rules(self, engine_speed_at_max_power):
        idle = self.idle_engine_speed[0]
        n_min_drive = idle + 0.125 * (engine_speed_at_max_power - idle)
        idle = {1: idle, 2: idle * 0.9}
        self.min_gear_vel = {
            g: idle.get(g, n_min_drive) * self.vsr[g] for g in self.gears
        }
        self.pipe.append(self.correct_driveability_rules)
        self.next_gears = []

    # noinspection PyUnresolvedReferences
    def correct_driveability_rules(
            self, gear, i, gears, times, velocities, accelerations,
            motive_powers, engine_temperatures, next_gear):
        if motive_powers is None:
            return gear
        if len(times) > len(motive_powers):
            times = times[:-1]
        if i == 0:
            pg = gear
        else:
            pg = gears[max(0, i - 1)]  # Previous gear.

        power = motive_powers[i]  # Current power.
        t0 = times[i]

        for j, (g, t) in enumerate(self.next_gears):
            if t0 <= t:
                gear = g
                self.next_gears = self.next_gears[j:]
                break
        else:
            self.next_gears = []

        # noinspection PyShadowingNames
        def get_next(k, g):
            t0 = times[k]
            for j, v in enumerate(self.next_gears):
                if t0 <= v[0]:
                    g = v[1]
                    break
            else:
                g = next_gear(g, k)
            # 4.3
            if g and motive_powers[k] < 0 and \
                    self.min_gear_vel[g] > velocities[k]:
                return 0
            return g

        # noinspection PyShadowingNames
        def up_clip(arr, k):
            return arr[min(i + 1, len(arr) - 1)]

        # 4.a During accelerations a gear have to last at least 2 seconds.
        if power > 0 or up_clip(velocities, i + 1) >= velocities[i]:
            b = [g == pg for g in gears[np.searchsorted(times, times[i] - 2):i]]
            gear = min(gear, pg + int(b and all(b) or 0))

        # 4.b
        if gear > 1 and power > 0:
            if pg < gear or motive_powers[max(i - 1, 0)] <= 0:
                g = gear
                t0 = times[i] + 10
                j = i + 1
                gen = enumerate(zip(times[j:], motive_powers[j:]), j)
                while True:
                    try:
                        k, (t, p) = next(gen)
                    except StopIteration:
                        break
                    if p <= 0:
                        break
                    g = get_next(k, g)
                    if g < gear:
                        if t <= t0:
                            gear = g
                        else:
                            t0 += 10
                            while True:
                                try:
                                    k, (t, p) = next(gen)
                                except StopIteration:
                                    break
                                if t > t0 or p <= 0:
                                    break
                                g = get_next(k, g)
                                if g < gear:
                                    gear = g
                                    break
                        break

            if pg > gear:
                g = gear
                t0 = times[i] + 10
                j = i + 1
                gen = enumerate(zip(times[j:], motive_powers[j:]), j)
                while True:
                    try:
                        k, (t, p) = next(gen)
                    except StopIteration:
                        break
                    g = get_next(k, g)
                    if g < pg:
                        break
                    if t > t0 or p <= 0:
                        gear = pg
                        break

        # 4.c, 4.d
        if pg and pg < gear:
            if power < 0 or up_clip(motive_powers, i + 1) < 0:  # 4.d
                gear = pg
            else:  # 4.c
                g = gear
                t0 = times[i] + 5.01
                j = i + 1
                gen = enumerate(times[j:], j)
                while True:
                    try:
                        k, t = next(gen)
                    except StopIteration:
                        break
                    if t > t0:
                        break
                    g = get_next(k, g)
                    if g < gear:
                        gear = max(pg, g)
                        break
        # 4.e
        if gear == 1 and pg == 2 and power < 0:
            for k, p in enumerate(motive_powers[i + 1:], i + 1):
                if p >= 0:
                    if velocities[k] <= 1:
                        gear = 2
                    break

        # 4.e
        if gear and power < 0 and self.min_gear_vel[gear] > velocities[i]:
            gear = 0

        # 4.f
        if gear and power < 0 and pg > gear:
            j, g = i - 1, gear
            t0 = times[max(j, 0)]
            b = [pg == g for g in gears[np.searchsorted(times[:j], t0 - 3):j]]
            if b and all(b):
                t1, t2, t3 = times[i], t0 + 2, t0 + 5
                flag = gear, t1 + 3
                j = i + 1
                gen = enumerate(zip(times[j:], motive_powers[j:]), j)
                for k, (t, p) in gen:
                    if not (p <= 0 or t <= t2):
                        break
                    g = get_next(k, g)
                    if g != flag[0]:
                        flag = g, t + 3

                    if t >= flag[1]:
                        if flag[0] < gear:
                            gear = 0
                            self.next_gears = [(0, t0 + 1), (flag[0], t0 + 2)]
                        break
                    elif t > t3:
                        break
                    elif t > t2 and 2 <= pg - g <= 3:
                        gear = 0
                        self.next_gears = [(0, t0 + 1), (flag[0], t0 + 2)]
                        break

                if not gear and pg > flag[0] + 1:
                    v, g0, r = None, g, flag[0] - 1
                    t1, t2 = t0 + 2, t0 + 5
                    for k, (t, p) in gen:
                        g = get_next(k, g)
                        if p > 0 or g > g0:
                            break
                        g0 = g
                        if v is None and t >= t1:
                            v = next_gear(g, k) - g <= 2
                        if t >= t2:
                            if g <= r:
                                if v:
                                    self.next_gears = [(0, t0 + 1), (r, t0 + 4)]
                                else:
                                    self.next_gears = [(0, t0 + 2), (g, t2)]
                            break

        # 4.f
        if power < 0 and gear and (pg > gear or pg == 0):
            j, g0 = i + 1, gear
            t0 = times[max(i - 1, 0)] + 2
            gen = enumerate(zip(times[j:], motive_powers[j:]), j)
            for k, (t, p) in gen:
                if p > 0:
                    break
                g1 = get_next(k, g0)
                if g1 == g0 and (g0 == 1 or t <= t0):
                    g0 = g1
                    continue
                if velocities[k] < 1:
                    self.next_gears = [(0, t)]
                    gear = 0
                elif not g1 and gear == 1:
                    g0 = g1
                    continue
                break

        # 3.2
        j = i + np.searchsorted(times[i:], times[i] + 1)
        if not gear and up_clip(velocities, j + 1) > up_clip(velocities, j):
            gear = self.min_gear

        return gear

    def __call__(self, gear, i, gears, times, velocities, accelerations,
                 motive_powers, engine_temperatures, next_gear):
        for f in self.pipe:
            gear = f(
                gear, i, gears, times, velocities, accelerations, motive_powers,
                engine_temperatures, next_gear
            )
        return gear


def _upgrade_gsm(gsm, velocity_speed_ratios, cycle_type):
    from .cmv import CMV
    from .gspv import GSPV
    from .dtgs import DTGS
    from .core import GSMColdHot
    if isinstance(gsm, (CMV, DTGS)):
        gsm = copy.deepcopy(gsm).convert(velocity_speed_ratios)
        if cycle_type == 'NEDC':
            if isinstance(gsm, MVL):
                par = dfl.functions.correct_constant_velocity
                gsm.correct_constant_velocity(
                    up_cns_vel=par.CON_VEL_DN_SHIFT,
                    dn_cns_vel=par.CON_VEL_UP_SHIFT
                )
            elif isinstance(gsm, CMV) and not isinstance(gsm, GSPV):
                par = dfl.functions.correct_constant_velocity
                gsm.correct_constant_velocity(
                    up_cns_vel=par.CON_VEL_UP_SHIFT,
                    up_window=par.VEL_UP_WINDOW,
                    up_delta=par.DV_UP_SHIFT, dn_cns_vel=par.CON_VEL_DN_SHIFT,
                    dn_window=par.VEL_DN_WINDOW, dn_delta=par.DV_DN_SHIFT
                )
    elif isinstance(gsm, GSMColdHot):
        gsm = copy.deepcopy(gsm)
        for k, v in gsm.items():
            gsm[k] = _upgrade_gsm(v, velocity_speed_ratios, cycle_type)

    return gsm


def correct_gear_v0(
        cycle_type, velocity_speed_ratios, mvl, idle_engine_speed,
        full_load_curve, max_velocity_full_load_correction=float('inf'),
        plateau_acceleration=float('inf')):
    """
    Returns a function to correct the gear predicted according to
    :func:`correct_gear_mvl` and :func:`correct_gear_full_load`.

    :param cycle_type:
        Cycle type (WLTP or NEDC).
    :type cycle_type: str

    :param velocity_speed_ratios:
        Constant velocity speed ratios of the gear box [km/(h*RPM)].
    :type velocity_speed_ratios: dict[int | float]

    :param mvl:
        Matrix velocity limits (upper and lower bound) [km/h].
    :type mvl: MVL

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :param full_load_curve:
        Vehicle full load curve.
    :type full_load_curve: function

    :param max_velocity_full_load_correction:
        Maximum velocity to apply the correction due to the full load curve.
    :type max_velocity_full_load_correction: float

    :param plateau_acceleration:
        Maximum acceleration to be at constant velocity [m/s2].
    :type plateau_acceleration: float

    :return:
        A function to correct the predicted gear.
    :rtype: callable
    """

    mvl = _upgrade_gsm(mvl, velocity_speed_ratios, cycle_type)
    mvl.plateau_acceleration = plateau_acceleration

    correct_gear = CorrectGear(velocity_speed_ratios, idle_engine_speed)
    correct_gear.fit_correct_gear_mvl(mvl)
    correct_gear.fit_correct_gear_full_load(
        full_load_curve, max_velocity_full_load_correction
    )
    correct_gear.fit_basic_correct_gear()

    return correct_gear


dsp.add_data(
    data_id='max_velocity_full_load_correction',
    default_value=dfl.values.max_velocity_full_load_correction
)

dsp.add_function(
    function=sh.add_args(correct_gear_v0),
    inputs=[
        'fuel_saving_at_strategy', 'cycle_type', 'velocity_speed_ratios', 'MVL',
        'idle_engine_speed', 'full_load_curve',
        'max_velocity_full_load_correction', 'plateau_acceleration'
    ],
    outputs=['correct_gear'],
    input_domain=co2_utl.check_first_arg
)


def correct_gear_v1(
        cycle_type, velocity_speed_ratios, mvl, idle_engine_speed,
        plateau_acceleration=float('inf')):
    """
    Returns a function to correct the gear predicted according to
    :func:`correct_gear_mvl`.

    :param cycle_type:
        Cycle type (WLTP or NEDC).
    :type cycle_type: str

    :param velocity_speed_ratios:
        Constant velocity speed ratios of the gear box [km/(h*RPM)].
    :type velocity_speed_ratios: dict[int | float]

    :param mvl:
        Matrix velocity limits (upper and lower bound) [km/h].
    :type mvl: OrderedDict

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :param plateau_acceleration:
        Maximum acceleration to be at constant velocity [m/s2].
    :type plateau_acceleration: float

    :return:
        A function to correct the predicted gear.
    :rtype: callable
    """

    mvl = _upgrade_gsm(mvl, velocity_speed_ratios, cycle_type)
    mvl.plateau_acceleration = plateau_acceleration

    correct_gear = CorrectGear(velocity_speed_ratios, idle_engine_speed)
    correct_gear.fit_correct_gear_mvl(mvl)
    correct_gear.fit_basic_correct_gear()

    return correct_gear


dsp.add_function(
    function=sh.add_args(correct_gear_v1),
    inputs=[
        'fuel_saving_at_strategy', 'cycle_type', 'velocity_speed_ratios', 'MVL',
        'idle_engine_speed', 'plateau_acceleration'
    ],
    outputs=['correct_gear'],
    weight=50,
    input_domain=co2_utl.check_first_arg
)


@sh.add_function(dsp, inputs_kwargs=True, outputs=['correct_gear'], weight=50)
def correct_gear_v2(
        velocity_speed_ratios, idle_engine_speed, full_load_curve,
        max_velocity_full_load_correction=float('inf')):
    """
    Returns a function to correct the gear predicted according to
    :func:`correct_gear_full_load`.

    :param velocity_speed_ratios:
        Constant velocity speed ratios of the gear box [km/(h*RPM)].
    :type velocity_speed_ratios: dict[int | float]

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :param full_load_curve:
        Vehicle full load curve.
    :type full_load_curve: function

    :param max_velocity_full_load_correction:
        Maximum velocity to apply the correction due to the full load curve.
    :type max_velocity_full_load_correction: float

    :return:
        A function to correct the predicted gear.
    :rtype: callable
    """

    correct_gear = CorrectGear(velocity_speed_ratios, idle_engine_speed)
    correct_gear.fit_correct_gear_full_load(
        full_load_curve, max_velocity_full_load_correction
    )
    correct_gear.fit_basic_correct_gear()

    return correct_gear


@sh.add_function(dsp, outputs=['correct_gear'], weight=100)
def correct_gear_v3(velocity_speed_ratios, idle_engine_speed):
    """
    Returns a function that does not correct the gear predicted.

    :param velocity_speed_ratios:
        Constant velocity speed ratios of the gear box [km/(h*RPM)].
    :type velocity_speed_ratios: dict[int | float]

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :return:
        A function to correct the predicted gear.
    :rtype: callable
    """

    correct_gear = CorrectGear(velocity_speed_ratios, idle_engine_speed)
    correct_gear.fit_basic_correct_gear()
    return correct_gear


@sh.add_function(dsp, outputs=['specific_gear_shifting'])
def default_specific_gear_shifting():
    """
    Returns the default value of specific gear shifting [-].

    :return:
        Specific gear shifting model.
    :rtype: str
    """
    return dfl.functions.default_specific_gear_shifting.SPECIFIC_GEAR_SHIFTING


# noinspection PyMissingOrEmptyDocstring
def at_domain(method):
    # noinspection PyMissingOrEmptyDocstring
    def domain(kwargs):
        return kwargs.get('specific_gear_shifting') in ('ALL', method)

    return domain


dsp.add_dispatcher(
    include_defaults=True,
    dsp_id='cmv_model',
    dsp=_cmv,
    input_domain=at_domain('CMV'),
    inputs=(
        'velocity_speed_ratios', 'engine_speeds_out', 'motive_powers', 'gears',
        'correct_gear', 'stop_velocity', 'accelerations', 'CMV', 'gear_filter',
        'velocities', 'cycle_type', 'times', {'specific_gear_shifting': sh.SINK}
    ),
    outputs=('CMV', 'gears')
)

dsp.add_dispatcher(
    include_defaults=True,
    dsp_id='cmv_ch_model',
    input_domain=at_domain('CMV_Cold_Hot'),
    dsp=_cmv_ch,
    inputs=(
        'CMV_Cold_Hot', 'accelerations', 'correct_gear', 'cycle_type',
        'engine_speeds_out', 'gear_filter', 'gears', 'stop_velocity',
        'time_cold_hot_transition', 'times', 'velocities', 'motive_powers',
        'velocity_speed_ratios', {'specific_gear_shifting': sh.SINK}
    ),
    outputs=('CMV_Cold_Hot', 'gears')
)

dsp.add_data(
    'use_dt_gear_shifting', dfl.values.use_dt_gear_shifting,
    description='If to use decision tree classifiers to predict gears.'
)


# noinspection PyMissingOrEmptyDocstring
def dt_domain(method):
    # noinspection PyMissingOrEmptyDocstring
    def domain(kw):
        s = kw.get('specific_gear_shifting')
        return s == method or (kw.get('use_dt_gear_shifting') and s == 'ALL')

    return domain


dsp.add_dispatcher(
    include_defaults=True,
    dsp_id='dtgs_model',
    input_domain=dt_domain('DTGS'),
    dsp=_dtgs,
    inputs=(
        'engine_temperatures', 'gear_filter', 'gears', 'motive_powers', 'times',
        'velocity_speed_ratios', 'velocities', 'accelerations', 'correct_gear',
        'DTGS', {
            'specific_gear_shifting': sh.SINK, 'use_dt_gear_shifting': sh.SINK
        }
    ),
    outputs=('DTGS', 'gears')
)

dsp.add_dispatcher(
    include_defaults=True,
    dsp_id='gspv_model',
    dsp=_gspv,
    input_domain=at_domain('GSPV'),
    inputs=(
        'GSPV', 'accelerations', 'correct_gear', 'cycle_type', 'velocities',
        'gear_filter', 'gears', 'motive_powers', 'stop_velocity', 'times',
        'velocity_speed_ratios', {'specific_gear_shifting': sh.SINK}
    ),
    outputs=('GSPV', 'gears')
)

dsp.add_dispatcher(
    include_defaults=True,
    dsp_id='gspv_ch_model',
    dsp=_gspv_ch,
    input_domain=at_domain('GSPV_Cold_Hot'),
    inputs=(
        'GSPV_Cold_Hot', 'accelerations', 'correct_gear', 'cycle_type', 'times',
        'gear_filter', 'gears', 'motive_powers', 'stop_velocity', 'velocities',
        'time_cold_hot_transition', 'velocity_speed_ratios',
        {'specific_gear_shifting': sh.SINK}
    ),
    outputs=('GSPV_Cold_Hot', 'gears')
)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the mechanic of the torque converter.
"""
import numpy as np
import schedula as sh
from co2mpas.defaults import dfl

dsp = sh.BlueDispatcher(
    name='Torque_converter', description='Models the torque converter.'
)


@sh.add_function(dsp, outputs=['clutch_window'])
def default_clutch_window():
    """
    Returns a default clutching time window [s] for a generic clutch.

    :return:
        Clutching time window [s].
    :rtype: tuple
    """
    return dfl.functions.default_clutch_window.clutch_window


@sh.add_function(dsp, inputs_kwargs=True, outputs=['k_factor_curve'])
def define_k_factor_curve(stand_still_torque_ratio=1.0, lockup_speed_ratio=0.0):
    """
    Defines k factor curve.

    :param stand_still_torque_ratio:
        Torque ratio when speed ratio==0.

        .. note:: The ratios are defined as follows:

           - Torque ratio = `gear box torque` / `engine torque`.
           - Speed ratio = `gear box speed` / `engine speed`.
    :type stand_still_torque_ratio: float

    :param lockup_speed_ratio:
        Minimum speed ratio where torque ratio==1.

        ..note::
            torque ratio==1 for speed ratio > lockup_speed_ratio.
    :type lockup_speed_ratio: float

    :return:
        k factor curve.
    :rtype: callable
    """
    from scipy.interpolate import InterpolatedUnivariateSpline
    if lockup_speed_ratio == 0:
        x = [0, 1]
        y = [1, 1]
    elif lockup_speed_ratio == 1:
        x = [0, 1]
        y = [stand_still_torque_ratio, 1]
    else:
        x = [0, lockup_speed_ratio, 1]
        y = [stand_still_torque_ratio, 1, 1]

    return InterpolatedUnivariateSpline(x, y, k=1)


@sh.add_function(dsp, outputs=['k_factor_curve'], weight=2)
def default_tc_k_factor_curve():
    """
    Returns a default k factor curve for a generic torque converter.

    :return:
        k factor curve.
    :rtype: callable
    """
    from co2mpas.defaults import dfl
    par = dfl.functions.default_tc_k_factor_curve
    a = par.STAND_STILL_TORQUE_RATIO, par.LOCKUP_SPEED_RATIO
    return define_k_factor_curve(*a)


@sh.add_function(
    dsp, outputs=['m1000_curve_ratios', 'm1000_curve_norm_torques'], weight=2
)
def default_tc_normalized_m1000_curve():
    """
    Returns default `m1000_curve_ratios` and `m1000_curve_norm_torques`.

    :return:
        Speed ratios and normalized torques of m1000 curve.
    :rtype: tuple[numpy.array]
    """
    from co2mpas.defaults import dfl
    curve = dfl.functions.default_tc_normalized_m1000_curve.curve
    return np.array(curve['x']), np.array(curve['y'])


# noinspection PyPep8Naming
@sh.add_function(dsp, outputs=['normalized_VDI253_model'], weight=2)
def define_normalized_VDI253_model(
        m1000_curve_ratios, m1000_curve_norm_torques, idle_engine_speed,
        engine_max_speed, k_factor_curve):
    """
    Defines normalized VDI253 model function.

    :param m1000_curve_ratios:
        Speed ratios of m1000 curve [-].
    :type m1000_curve_ratios: numpy.array

    :param m1000_curve_norm_torques:
        Normalized torques of m1000 curve [-].
    :type m1000_curve_norm_torques: numpy.array

    :param idle_engine_speed:
        Idle engine speed and its standard deviation [RPM].
    :type idle_engine_speed: (float, float)

    :param engine_max_speed:
        Maximum allowed engine speed [RPM].
    :type engine_max_speed: float

    :param k_factor_curve:
        k factor curve.
    :type k_factor_curve: callable

    :return:
        Normalized VDI253 model function.
    :rtype: scipy.interpolate.LinearNDInterpolator
    """
    from scipy.interpolate import interp1d, LinearNDInterpolator
    maximum_ratio, idle = np.max(m1000_curve_ratios), idle_engine_speed[0]
    eng_s = np.linspace(idle, engine_max_speed, 100)
    gb_s = np.linspace(0, engine_max_speed * maximum_ratio, 250)
    x, z = np.meshgrid(gb_s, eng_s)
    r = x / z
    b = r <= maximum_ratio
    x, z, r = x[b], z[b], r[b]
    func = interp1d(m1000_curve_ratios, m1000_curve_norm_torques, kind='cubic')
    tout = func(r) * k_factor_curve(r) * z ** 2
    return LinearNDInterpolator((x, tout), z)


# noinspection PyPep8Naming
@sh.add_function(dsp, outputs=['m1000_curve_factor'])
def calibrate_m1000_curve_factor(
        full_load_curve, normalized_VDI253_model, clutch_phases,
        engine_speeds_out_hot, gear_box_speeds_in, gear_box_torques_in,
        clutch_tc_speeds_delta):
    """
    Calibrate the rescaling factor of m1000 curve [N*m/1e6].

    :param full_load_curve:
        Vehicle full load curve.
    :type full_load_curve: function

    :param normalized_VDI253_model:
        Normalized VDI253 model function.
    :type normalized_VDI253_model: scipy.interpolate.LinearNDInterpolator

    :param clutch_phases:
        When the clutch is active [-].
    :type clutch_phases: numpy.array

    :param engine_speeds_out_hot:
        Engine speed at hot condition [RPM].
    :type engine_speeds_out_hot: numpy.array

    :param gear_box_speeds_in:
        Gear box speed vector [RPM].
    :type gear_box_speeds_in: numpy.array

    :param gear_box_torques_in:
        Torque required vector [N*m].
    :type gear_box_torques_in: numpy.array

    :param clutch_tc_speeds_delta:
        Engine speed delta due to the clutch or torque converter [RPM].
    :type clutch_tc_speeds_delta: numpy.array

    :return:
        Rescaling factor of m1000 curve [N*m/1e6].
    :rtype: float
    """
    if clutch_phases.sum() <= 10:
        return sh.NONE
    from co2mpas.utils import mae
    from scipy.optimize import fmin

    # noinspection PyUnresolvedReferences
    es, gbs, gbt, predict, ds = (
        engine_speeds_out_hot[clutch_phases], gear_box_speeds_in[clutch_phases],
        gear_box_torques_in[clutch_phases], normalized_VDI253_model.predict,
        clutch_tc_speeds_delta[clutch_phases]

    )

    def _err(factor):
        e = mae(ds, np.nan_to_num(predict((gbs, gbt / factor)) - es))
        return np.float32(e)

    return fmin(_err, default_m1000_curve_factor(full_load_curve))


@sh.add_function(dsp, outputs=['m1000_curve_factor'], weight=1000)
def default_m1000_curve_factor(full_load_curve):
    """
    Returns the default value of the rescaling factor of m1000 curve [N*m].

    :param full_load_curve:
        Vehicle full load curve.
    :type full_load_curve: function

    :return:
        Rescaling factor of m1000 curve [N*m/1e6].
    :rtype: float
    """
    from ..wheels import calculate_wheel_torques
    return calculate_wheel_torques(full_load_curve(1000), 1000) / 1e6


# noinspection PyPep8Naming
@sh.add_function(dsp, outputs=['torque_converter_speed_model'])
def define_tc_speed_model(
        normalized_VDI253_model, m1000_curve_factor, idle_engine_speed):
    """
    Define torque converter speed model.

    :param normalized_VDI253_model:
        Normalized VDI253 model function.
    :type normalized_VDI253_model: scipy.interpolate.LinearNDInterpolator

    :param m1000_curve_factor:
        Rescaling factor of m1000 curve [N*m/1e6].
    :type m1000_curve_factor: float

    :param idle_engine_speed:
        Idle engine speed and its standard deviation [RPM].
    :type idle_engine_speed: (float, float)

    :return:
        Torque converter speed model.
    :rtype: callable
    """

    # noinspection PyMissingOrEmptyDocstring,PyUnusedLocal
    def model(times, **kwargs):
        gbs, gbt = kwargs['gear_box_speeds_in'], kwargs['gear_box_torques_in']
        es = normalized_VDI253_model((gbs, gbt / m1000_curve_factor))
        return np.nan_to_num(es - np.maximum(gbs, idle_engine_speed[0]))

    return model



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
It provides validation functions and the validation model `dsp`.

Sub-Modules:

.. currentmodule:: co2mpas.core.load.validate

.. autosummary::
    :nosignatures:
    :toctree: validate/

    hard
"""
import logging
import schedula as sh

log = logging.getLogger(__name__)
dsp = sh.BlueDispatcher(
    name='validate_data', description='Validates input data.'
)


def _add_validated_input(data, validate, keys, value, errors):
    from schema import SchemaError
    try:
        try:
            k, v = validate(keys[-1], value)
        except TypeError:
            k, v = next(iter(validate({keys[-1]: value}).items()))
        if v is not sh.NONE:
            data[k] = v
            return v
    except SchemaError as ex:
        sh.get_nested_dicts(errors, *keys[:-1])[keys[-1]] = ex
    return sh.NONE


def _validate_base_with_schema(data, depth=4):
    from ..schema import define_data_validation
    inputs, errors, validate = {}, {}, define_data_validation()
    for k, v in sorted(sh.stack_nested_keys(data, depth=depth)):
        d = sh.get_nested_dicts(inputs, *k[:-1])
        _add_validated_input(d, validate, k, v, errors)

    return inputs, errors


def _log_errors_msg(errors):
    if errors:
        msg = ['\nInput cannot be parsed, due to:']
        for k, v in sh.stack_nested_keys(errors, depth=4):
            msg.append('{} in {}: {}'.format(k[-1], '/'.join(k[:-1]), v))
        log.error('\n  '.join(msg))
        return True
    return False


def _mode_parser(
        type_approval_mode, declaration_mode, hard_validation, inputs, errors,
        input_type=None):

    if type_approval_mode or declaration_mode:
        from co2mpas_dice.declaration import declaration_validation
        inputs, errors = declaration_validation(
            type_approval_mode, inputs, errors, input_type=input_type
        )

    if hard_validation:
        from .hard import hard_validation as validation
        inputs, errors = validation(inputs, errors)

    return inputs, errors


_kw = dict(inputs_kwargs=True, inputs_defaults=True)


@sh.add_function(dsp, outputs=['validated_base'], **_kw)
def validate_base(
        input_type, base=None, declaration_mode=False, hard_validation=False,
        enable_selector=False, type_approval_mode=False):
    """
    Validate base data.

    :param input_type:
        Type of file input.
    :type input_type: str

    :param base:
        Base data.
    :type base: dict

    :param declaration_mode:
        Use only the declaration data.
    :type declaration_mode: bool

    :param hard_validation:
        Add extra data validations.
    :type hard_validation: bool

    :param enable_selector:
        Enable the selection of the best model to predict both H/L cycles.
    :type enable_selector: bool

    :param type_approval_mode:
        Is launched for TA?
    :type type_approval_mode: bool

    :return:
        Validated base data.
    :rtype: dict
    """
    i, e = _validate_base_with_schema(base or {})

    i, e = _mode_parser(
        type_approval_mode, declaration_mode, hard_validation, i, e, input_type
    )

    if _log_errors_msg(e):
        return sh.NONE
    i['enable_selector'] = enable_selector
    return {'.'.join(k): v for k, v in sh.stack_nested_keys(i, depth=3)}


@sh.add_function(dsp, outputs=['validated_meta'], **_kw)
def validate_meta(meta=None, hard_validation=False):
    """
    Validate meta data.

    :param meta:
        Meta data.
    :type meta: dict

    :param hard_validation:
        Add extra data validations.
    :type hard_validation: bool

    :return:
        Validated meta data.
    :rtype: dict
    """
    i, e = _validate_base_with_schema(meta or {}, depth=2)
    if hard_validation:
        from schema import SchemaError
        from .hard import _hard_validation
        for k, v in sorted(sh.stack_nested_keys(i, depth=1)):
            for c, msg in _hard_validation(v, 'meta'):
                sh.get_nested_dicts(e, *k)[c] = SchemaError([], [msg])

    if _log_errors_msg(e):
        return sh.NONE

    return i


@sh.add_function(dsp, outputs=['validated_dice'], **_kw)
def validate_dice(dice=None):
    """
    Validate DICE data.

    :param dice:
        DICE data.
    :type dice: dict

    :return:
        Validated DICE data.
    :rtype: dict
    """
    try:
        from co2mpas_dice.verify import validate_dice
        return validate_dice(dice)
    except ImportError:
        return dice or {}


@sh.add_function(dsp, outputs=['input_type'], **_kw)
def get_input_type(validated_dice):
    """
    Return the input type.

    :param validated_dice:
        Validated DICE data.
    :type validated_dice: dict

    :return:
        Type of file input.
    :rtype: str
    """
    return validated_dice.get('input_type')


@sh.add_function(dsp, outputs=['validated_flag'], **_kw)
def validate_flag(flag=None, declaration_mode=False, type_approval_mode=False):
    """
    Validate flags data.

    :param flag:
        Flags data.
    :type flag: dict

    :param declaration_mode:
        Use only the declaration data.
    :type declaration_mode: bool

    :param type_approval_mode:
        Is launched for TA?
    :type type_approval_mode: bool

    :return:
        Validated flags data.
    :rtype: dict
    """
    from ..schema import define_flags_validation
    inputs, errors, validate = {}, {}, define_flags_validation()
    for k, v in sorted((flag or {}).items()):
        _add_validated_input(inputs, validate, ('flag', k), v, errors)
    if declaration_mode or type_approval_mode:
        from co2mpas_dice.verify import verify_flag
        if not verify_flag(inputs):
            return sh.NONE
    if _log_errors_msg(errors):
        return sh.NONE
    return inputs


@sh.add_function(dsp, outputs=['validated_plan'], **_kw)
def validate_plan(
        input_type, plan=None, declaration_mode=False, hard_validation=False,
        type_approval_mode=False):
    """
    Validate plan data.

    :param input_type:
        Type of file input.
    :type input_type: str

    :param plan:
        Plan data.
    :type plan: dict

    :param declaration_mode:
        Use only the declaration data.
    :type declaration_mode: bool

    :param hard_validation:
        Add extra data validations.
    :type hard_validation: bool

    :param type_approval_mode:
        Is launched for TA?
    :type type_approval_mode: bool

    :return:
        Validated plan data.
    :rtype: dict
    """
    import pandas as pd
    if plan and declaration_mode:
        msg = 'Simulation plan cannot be executed in declaration mode!\n' \
              'If you want to execute it remove -DM or -TA from the cmd.'
        log.warning(msg)
        return []
    import os.path as osp
    from ..excel import _parse_values as parse_key
    from ..schema import define_data_validation as _schema

    validated_plan, e, keys = [], {}, {'id', 'base', 'run_base'}
    validate, add = _schema(), validated_plan.append

    for _, d in pd.DataFrame(plan or []).iterrows():
        d = dict(d.dropna(how='all'))
        d['base'] = osp.abspath(d['base'])
        i, data, p_id = {}, {}, 'plan id:{}'.format(d['id'])

        for k, v in parse_key(sh.selector(set(d) - keys, d), where='in plan'):
            k, inp = (p_id,) + k, sh.get_nested_dicts(i, *k[1:-1])
            v = _add_validated_input(inp, validate, k, v, e)
            if v is not sh.NONE:
                sh.get_nested_dicts(data, '.'.join(k[2:-1]))[k[-1]] = v

        e = _mode_parser(
            type_approval_mode, declaration_mode, hard_validation, i, e,
            input_type
        )[1]
        add(sh.combine_dicts({'data': data}, base=sh.selector(keys, d)))

    if _log_errors_msg(e):
        return sh.NONE

    return validated_plan


@sh.add_function(dsp, outputs=['verified'], **_kw)
def validation_status(
        validated_base, validated_flag, validated_meta, validated_dice,
        validated_plan, type_approval_mode=False):
    """
    Returns the validation status.

    :param validated_base:
        Validated base data.
    :type validated_base: dict

    :param validated_flag:
        Validated flags data.
    :type validated_flag: dict

    :param validated_meta:
        Validated meta data.
    :type validated_meta: dict

    :param validated_dice:
        Validated DICE data.
    :type validated_dice: dict

    :param validated_plan:
        Validated plan data.
    :type validated_plan: dict

    :param type_approval_mode:
        Is launched for TA?
    :type type_approval_mode: bool

    :return:
        Validation status.
    :rtype: bool
    """
    if type_approval_mode:
        from co2mpas_dice.verify import verify_data
        return verify_data({
            'base': validated_base,
            'flag': validated_flag,
            'meta': validated_meta,
            'dice': validated_dice,
            'plan': validated_plan
        })
    return True



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to calculate the WLTP theoretical velocities.
"""
import numpy as np
import schedula as sh
from co2mpas.defaults import dfl
from ...vehicle import dsp as _vehicle

dsp = sh.BlueDispatcher(
    name='WLTP velocities model',
    description='Returns the theoretical velocities of WLTP.'
)


@sh.add_function(
    dsp, outputs=['resistance_coeffs_regression_curves', 'wltc_data']
)
def get_dfl(base_model):
    """
    Gets default values from wltp base model.

    :param base_model:
        WLTP base model.
    :type base_model: dict

    :return:
        Default values from wltp base model.
    :rtype: list
    """
    params = base_model['params']
    keys = 'resistance_coeffs_regression_curves', 'wltc_data'
    return sh.selector(keys, params, output_type='list')


@sh.add_function(dsp, outputs=['road_loads'], weight=15)
def default_road_loads(
        vehicle_mass, resistance_coeffs_regression_curves):
    """
    Returns default road loads.

    :param vehicle_mass:
        Vehicle mass [kg].
    :type vehicle_mass: float

    :param resistance_coeffs_regression_curves:
        Regression curve coefficient to calculate the default road loads.
    :type resistance_coeffs_regression_curves: list[list[float]]

    :return:
        Cycle road loads [N, N/(km/h), N/(km/h)^2].
    :rtype: list, tuple
    """
    from wltp.experiment import calc_default_resistance_coeffs as func
    return func(vehicle_mass, resistance_coeffs_regression_curves)


@sh.add_function(dsp, outputs=['max_speed_velocity_ratio'])
def calculate_max_speed_velocity_ratio(speed_velocity_ratios):
    """
    Calculates the maximum speed velocity ratio of the gear box [h*RPM/km].

    :param speed_velocity_ratios:
        Speed velocity ratios of the gear box [h*RPM/km].
    :type speed_velocity_ratios: dict[int | float]

    :return:
        Maximum speed velocity ratio of the gear box [h*RPM/km].
    :rtype: float
    """

    return speed_velocity_ratios[max(speed_velocity_ratios)]


@sh.add_function(dsp, outputs=['max_velocity'])
def calculate_max_velocity(engine_speed_at_max_power, max_speed_velocity_ratio):
    """
    Calculates max vehicle velocity [km/h].

    :param engine_speed_at_max_power:
        Rated engine speed [RPM].
    :type engine_speed_at_max_power: float

    :param max_speed_velocity_ratio:
        Maximum speed velocity ratio of the gear box [h*RPM/km].
    :type max_speed_velocity_ratio: float

    :return:
        Max vehicle velocity [km/h].
    :rtype: float
    """

    return engine_speed_at_max_power / max_speed_velocity_ratio


@sh.add_function(dsp, outputs=['wltp_class'])
def calculate_wltp_class(
        wltc_data, engine_max_power, unladen_mass, max_velocity):
    """
    Calculates the WLTP vehicle class.

    :param wltc_data:
        WLTC data.
    :type wltc_data: dict

    :param engine_max_power:
        Maximum power [kW].
    :type engine_max_power: float

    :param unladen_mass:
        Unladen mass [kg].
    :type unladen_mass: float

    :param max_velocity:
        Max vehicle velocity [km/h].
    :type max_velocity: float

    :return:
        WLTP vehicle class.
    :rtype: str
    """
    from wltp.experiment import decideClass
    ratio = 1000.0 * engine_max_power / unladen_mass
    return decideClass(wltc_data, ratio, max_velocity)


@sh.add_function(dsp, outputs=['class_data'])
def get_class_data(wltc_data, wltp_class):
    """
    Returns WLTP class data.

    :param wltc_data:
        WLTC data.
    :type wltc_data: dict

    :param wltp_class:
        WLTP vehicle class.
    :type wltp_class: str

    :return:
        WLTP class data.
    :rtype: dict
    """

    return wltc_data['classes'][wltp_class]


@sh.add_function(dsp, outputs=['class_times', 'class_velocities'], weight=25)
def get_class_velocities(class_data):
    """
    Returns time and velocity profiles according to WLTP class data [s, km/h].

    :param class_data:
        WLTP class data.
    :type class_data: dict

    :return:
        Class time and velocity vectors [s, km/h].
    :rtype: tuple[numpy.array]
    """
    vel = np.asarray(class_data['cycle'], dtype=float)
    return np.arange(vel.shape[0], dtype=float), vel


i = ['vehicle_mass', 'road_loads', 'inertial_factor']
calculate_class_powers = sh.SubDispatchPipe(
    _vehicle,
    function_id='calculate_class_powers',
    inputs=['times', 'velocities'] + i,
    outputs=['motive_powers']
)
dsp.add_function(
    function=calculate_class_powers,
    inputs=['class_times', 'class_velocities'] + i,
    outputs=['class_powers']
)

dsp.add_data(
    'downscale_factor_threshold', dfl.values.downscale_factor_threshold
)


@sh.add_function(dsp, outputs=['downscale_factor'])
def calculate_downscale_factor(
        class_data, downscale_factor_threshold, max_velocity, engine_max_power,
        class_powers):
    """
    Calculates velocity downscale factor [-].

    :param class_data:
        WLTP class data.
    :type class_data: dict

    :param downscale_factor_threshold:
        Velocity downscale factor threshold [-].
    :type downscale_factor_threshold: float

    :param max_velocity:
        Max vehicle velocity [km/h].
    :type max_velocity: float

    :param engine_max_power:
        Maximum power [kW].
    :type engine_max_power: float

    :param class_powers:
        Class motive power [kW].
    :type class_powers: numpy.array

    :return:
        Velocity downscale factor [-].
    :rtype: float
    """
    from wltp.experiment import calcDownscaleFactor
    dsc_data = class_data['downscale']
    p_max_values = dsc_data['p_max_values']
    downsc_coeffs = dsc_data['factor_coeffs']
    dsc_v_split = dsc_data.get('v_max_split', None)
    downscale_factor = calcDownscaleFactor(
        class_powers, p_max_values, downsc_coeffs, dsc_v_split,
        engine_max_power, max_velocity, downscale_factor_threshold
    )
    return downscale_factor


@sh.add_function(dsp, outputs=['downscale_phases'])
def get_downscale_phases(class_data):
    """
    Returns downscale phases [s].

    :param class_data:
        WLTP class data.
    :type class_data: dict

    :return:
        Downscale phases [s].
    :rtype: list
    """
    return class_data['downscale']['phases']


@sh.add_function(dsp, outputs=['theoretical_velocities'])
def wltp_velocities(
        downscale_factor, class_times, class_velocities, downscale_phases,
        times):
    """
    Returns the downscaled velocity profile [km/h].

    :param downscale_factor:
        Velocity downscale factor [-].
    :type downscale_factor: float

    :param class_times:
        Class time vector [s].
    :type class_times: numpy.array

    :param class_velocities:
        Class velocity vector [km/h].
    :type class_velocities: numpy.array

    :param downscale_phases:
        Downscale phases [s].
    :type downscale_phases: list

    :param times:
        Time vector [s].
    :type times: numpy.array

    :return:
        Theoretical velocity vector [km/h].
    :rtype: numpy.array
    """

    if downscale_factor > 0:
        from wltp.experiment import downscaleCycle
        downscale_phases = np.searchsorted(times, downscale_phases)
        v = downscaleCycle(class_velocities, downscale_factor, downscale_phases)
    else:
        v = class_velocities

    n = int(np.ceil(times[-1] / class_times[-1]))
    t = np.cumsum(np.tile(np.ediff1d(class_times, to_begin=[0]), (n,)))
    t += class_times[0]
    return np.interp(times, t, np.tile(v, (n,)))


dsp.add_function(
    function_id='calculate_theoretical_motive_powers',
    function=calculate_class_powers,
    inputs=['times', 'theoretical_velocities'] + i,
    outputs=['theoretical_motive_powers']
)
dsp.add_function(
    function=sh.bypass,
    inputs=['theoretical_velocities'],
    outputs=['velocities']
)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Constants for the CO2MPAS physical model.
"""
import schedula as sh
import co2mpas.utils as co2_utl


#: Container of node default values.
# noinspection PyMissingOrEmptyDocstring,PyPep8Naming
class Values(co2_utl.Constants):
    #: Minimum distance between points for the elevation interpolation [m].
    minimum_elevation_distance = 30

    #: Is the vehicle plugin hybrid?
    is_plugin = False

    #: Is the vehicle serial hybrid?
    is_serial = False

    #: Apply RCB correction?
    rcb_correction = True

    #: Family correction factor for representative regional temperatures [-].
    atct_family_correction_factor = 1.0

    #: Drive battery technology type.
    drive_battery_technology = 'unknown'

    #: Belt efficiency [-].
    belt_efficiency = 0.8

    #: Starter efficiency [-].
    starter_efficiency = 0.7

    #: Tyre state (i.e., new or worm).
    tyre_state = 'new'

    #: Road state (i.e., dry, wet, rainfall, puddles, ice).
    road_state = 'dry'

    #: Number of engine cylinders [-].
    engine_n_cylinders = 4

    #: Default additive correction for vehicles with periodically regenerating
    #: systems [CO2g/km].
    ki_additive = 0

    #: Number of passengers including driver [-].
    n_passengers = 1

    #: Average passenger mass [kg].
    passenger_mass = 75

    #: Cargo mass [kg].
    cargo_mass = 0

    #: Fuel mass [kg].
    fuel_mass = 25

    #: Does the engine have selective catalytic reduction technology?
    has_selective_catalytic_reduction = False

    #: Does the engine have lean burn technology?
    has_lean_burn = False

    #: Does the gear box have some additional technology to heat up faster?
    has_gear_box_thermal_management = False

    #: Does the vehicle has periodically regenerating systems? [-].
    has_periodically_regenerating_systems = False

    #: Possible percentages of active cylinders [-].
    active_cylinder_ratios = (1.0,)

    #: Does the engine feature variable valve actuation? [-].
    engine_has_variable_valve_actuation = False

    #: NEDC cycle time [s].
    max_time_NEDC = 1180.0

    #: WLTP cycle time [s].
    max_time_WLTP = 1800.0

    #: Maximum velocity to consider the vehicle stopped [km/h].
    stop_velocity = 1.0 + 1.1920929e-07

    #: Maximum acceleration to be at constant velocity [m/s2].
    plateau_acceleration = 0.1 + 1.1920929e-07

    #: Does the vehicle have start/stop system?
    has_start_stop = True

    #: Does the engine have cylinder deactivation technology?
    engine_has_cylinder_deactivation = False

    #: Minimum vehicle engine speed [RPM].
    min_engine_on_speed = 10.0

    #: Minimum time of engine on after a start [s].
    min_time_engine_on_after_start = 4.0

    #: Time window used to apply gear change filters [s].
    change_gear_window_width = 4.0

    #: Service battery start window width [s].
    service_battery_start_window_width = 4.0

    #: Threshold vehicle velocity for gear correction due to full load curve
    #: [km/h].
    max_velocity_full_load_correction = 100.0

    #: Air temperature [°C].
    air_temperature = 20

    #: Atmospheric pressure [kPa].
    atmospheric_pressure = 101.325

    #: Has the vehicle a roof box? [-].
    has_roof_box = False

    #: Tyre class (C1, C2, and C3).
    tyre_class = 'C1'

    #: Angle slope [rad].
    angle_slope = 0.0

    #: A different preconditioning cycle was used for WLTP and NEDC?
    correct_f0 = False

    #: Final drive ratio [-].
    final_drive_ratio = 1.0

    #: Wheel drive (i.e., front, rear, front+rear).
    wheel_drive = 'front'

    #: Apply the eco-mode gear shifting?
    fuel_saving_at_strategy = True

    #: Cold and hot gear box reference temperatures [°C].
    gear_box_temperature_references = (40.0, 80.0)

    #: Constant torque loss factors due to engine auxiliaries [N/cc, N*m].
    auxiliaries_torque_loss_factors = (0.175, 0.2021)  # m, q

    #: Constant power loss due to engine auxiliaries [kW].
    auxiliaries_power_loss = 0.0213

    #: If the engine is equipped with any kind of charging.
    engine_is_turbo = True

    #: Start-stop activation time threshold [s].
    start_stop_activation_time = 30

    #: Standard deviation of idle engine speed [RPM].
    idle_engine_speed_std = 100.0

    #: Is an hot cycle?
    is_cycle_hot = False

    #: CO2 emission model params.
    co2_params = {}

    #: Enable the calculation of Willans coefficients for all phases?
    enable_phases_willans = False

    #: Enable the calculation of Willans coefficients for the cycle?
    enable_willans = False

    #: Alternator efficiency [-].
    alternator_efficiency = 0.67

    #: Time elapsed to turn on the engine with electric starter [s].
    delta_time_engine_starter = .5

    #: If to use decision tree classifiers to predict gears.
    use_dt_gear_shifting = False

    #: Does the vehicle have energy recuperation features?
    has_energy_recuperation = True

    #: A/T Time at cold hot transition phase [s].
    time_cold_hot_transition = 300.0

    #: Time frequency [1/s].
    time_sample_frequency = 1.0

    #: Initial temperature of the test cell of NEDC [°C].
    initial_temperature_NEDC = 25.0

    #: Initial temperature of the test cell of WLTP [°C].
    initial_temperature_WLTP = 23.0

    #: K1 NEDC parameter (first or second gear) [-].
    k1 = 1

    #: K2 NEDC parameter (first or second gear) [-].
    k2 = 2

    #: K5 NEDC parameter (first or second gear) [-].
    k5 = 2

    #: WLTP base model params.
    wltp_base_model = {}

    #: Velocity downscale factor threshold [-].
    downscale_factor_threshold = 0.01

    #: Empirical value in case of CVT [-].
    tyre_dynamic_rolling_coefficient = 3.05 / 3.14


#: Container of internal function parameters.
# noinspection PyMissingOrEmptyDocstring,PyPep8Naming
class Functions(co2_utl.Constants):
    ENABLE_ALL_FUNCTIONS = False

    class default_after_treatment_warm_up_duration(co2_utl.Constants):
        #: After treatment warm up duration for conventional vehicles [s].
        duration = 60

    class default_after_treatment_cooling_duration(co2_utl.Constants):
        #: After treatment cooling duration for conventional vehicles [s].
        duration = float('inf')

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class parse_solution(co2_utl.Constants):
        #: Use all calibration outputs as relative prediction targets.
        CALIBRATION_AS_TARGETS = False

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class select_best_model(co2_utl.Constants):
        #: Model selector mapping.
        MAP = {None: {
            'wltp_h': ('nedc_h', 'wltp_h'),
            'wltp_l': ('nedc_l', 'wltp_l'),
            'wltp_m': ('wltp_m',)
        }}

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class calculate_cylinder_deactivation_valid_phases(co2_utl.Constants):
        #: Engine inertia  [kW].
        LIMIT = 0.0001

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class calculate_drive_battery_n_parallel_cells_v1(co2_utl.Constants):
        #: Voltage for calculating the number of parallel battery cells [V].
        reference_volt = {
            'unknown': 2.9,
            'NiMH': 1.1,
            'Li-NCA (Li-Ni-Co-Al)': 2.9,
            'Li-NCM (Li-Ni-Mn-Co)': 2.9,
            'Li-MO (Li-Mn)': 2.9,
            'Li-FP (Li-Fe-P)': 2.4,
            'Li-TO (Li-Ti)': 1.7
        }  # source: batteryuniversity.com

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class default_planetary_mean_efficiency(co2_utl.Constants):
        #: Planetary mean efficiency [-].
        efficiency = .97

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class default_planetary_ratio(co2_utl.Constants):
        #: Fundamental planetary speed ratio [-].
        ratio = 2.6

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class identify_after_treatment_warm_up_phases(co2_utl.Constants):
        #: After treatment cooling time [s].
        cooling_time = 400

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class define_service_battery_electric_powers_supply_threshold(
        co2_utl.Constants):
        #: Minimum SOC variation to define service battery charging status [%].
        min_soc = 0.1

        #: Maximum allowed negative current for the service battery being
        #: considered not charging [A].
        min_current = -1.0

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class default_final_drive_efficiency(co2_utl.Constants):
        #: Formula for the default final drive efficiency [function].
        formula = '1 - (n_wheel_drive - 2) / 100'

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class predict_clutch_tc_speeds_delta(co2_utl.Constants):
        #: Enable prediction of clutch or torque converter speeds delta?
        ENABLE = False

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class default_tc_normalized_m1000_curve(co2_utl.Constants):
        #: Normalized m1000 curve [-].
        curve = dict(
            x=[
                0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.65, 0.7, 0.75, 0.8, 0.87,
                0.9, 0.95, 1, 1.1, 1.222, 1.375, 1.571, 1.833, 2.2, 2.5, 3, 3.5,
                4, 4.5, 5
            ],
            y=[
                1, 0.97, 0.93, 0.9, 0.87, 0.83, 0.8, 0.75, 0.7, 0.65, 0.6, 0.55,
                0.5, 0.25, 0, -0.099, -0.198, -0.336, -0.535, -0.828, -1.306,
                -1.781, -2.772, -4.071, -5.746, -7.861, -10.480
            ]
        )

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class default_n_wheel(co2_utl.Constants):
        #: Total number of wheels [-].
        n_wheel = 4

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class default_static_friction(co2_utl.Constants):
        #: Static friction coefficient [-].
        coeff = dict(
            new=dict(dry=.85, wet=.65, rainfall=.55, puddles=.5, ice=.1),
            worm=dict(dry=1, wet=.5, rainfall=.4, puddles=.25, ice=.1)
        )

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class calculate_velocities(co2_utl.Constants):
        #: Time window for the moving average of obd velocities [s].
        dt_window = 5

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class calculate_engine_temperature_derivatives(co2_utl.Constants):
        #: Derivative spacing [s].
        dx = 4

        #: Degree of the smoothing spline [-].
        order = 7

        #: Time window for smoother [s].
        tw = 20

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class default_clutch_window(co2_utl.Constants):
        #: Clutching time window [s].
        clutch_window = (0, 0.95384615)

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class define_fuel_type_and_is_hybrid(co2_utl.Constants):
        #: Is the vehicle hybrid?
        is_hybrid = {
            # 0: False,  # Not available
            1: False,  # Gasoline
            2: False,  # Methanol
            3: False,  # Ethanol
            4: False,  # Diesel
            5: False,  # LPG
            6: False,  # CNG
            7: False,  # Propane
            8: True,  # Electric
            9: False,  # Bifuel running Gasoline
            10: False,  # Bifuel running Methanol
            11: False,  # Bifuel running Ethanol
            12: False,  # Bifuel running LPG
            13: False,  # Bifuel running CNG
            14: False,  # Bifuel running Propane
            15: True,  # Bifuel running Electricity
            16: True,  # Bifuel running electric and combustion engine
            17: True,  # Hybrid gasoline
            18: True,  # Hybrid Ethanol
            19: True,  # Hybrid Diesel
            20: True,  # Hybrid Electric
            21: True,  # Hybrid running electric and combustion engine
            22: True,  # Hybrid Regenerative
            23: False,  # Bifuel running diesel
        }
        #: The vehicle fuel type.
        fuel_type = {
            # 0: 'gasoline',  # Not available
            1: 'gasoline',  # Gasoline
            2: 'methanol',  # Methanol
            3: 'ethanol',  # Ethanol
            4: 'diesel',  # Diesel
            5: 'LPG',  # LPG
            6: 'NG',  # CNG
            7: 'propane',  # Propane
            8: None,  # Electric
            9: 'gasoline',  # Bifuel running Gasoline
            10: 'methanol',  # Bifuel running Methanol
            11: 'ethanol',  # Bifuel running Ethanol
            12: 'LPG',  # Bifuel running LPG
            13: 'NG',  # Bifuel running CNG
            14: 'propane',  # Bifuel running Propane
            15: None,  # Bifuel running Electricity
            16: 'gasoline',  # Bifuel running electric and combustion engine
            17: 'gasoline',  # Hybrid gasoline
            18: 'ethanol',  # Hybrid Ethanol
            19: 'diesel',  # Hybrid Diesel
            20: None,  # Hybrid Electric
            21: 'gasoline',  # Hybrid running electric and combustion engine
            22: None,  # Hybrid Regenerative
            23: 'diesel',  # Bifuel running diesel
        }

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class CMV(co2_utl.Constants):
        #: Enable optimization loop?
        ENABLE_OPT_LOOP = False

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class default_start_stop_activation_time(co2_utl.Constants):
        #: Enable function?
        ENABLE = False

        #: Start-stop activation time threshold [s].
        threshold = 30

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class calculate_last_gear_box_ratio_v1(co2_utl.Constants):
        #: Maximum admissible ratio for the last gear [-].
        MAX_RATIO = 2

        #: Minimum admissible ratio for the last gear [-].
        MIN_RATIO = 0.2

        #: Delta ratio for calculating the last gear [-].
        DELTA_RATIO = 0.1

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class calculate_maximum_velocity(co2_utl.Constants):
        #: Maximum admissible velocity for the vehicle maximum velocity [km/h].
        MAX_VEL = 1000

        #: Minimum admissible velocity for the vehicle maximum velocity [km/h].
        MIN_VEL = 1

        #: Delta ratio for calculating the vehicle maximum velocity [km/h].
        DELTA_VEL = 1

        #: Full load curve percentage fro calculating the available power [-].
        PREC_FLC = 0.9

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class MGS(co2_utl.Constants):
        #: Maximum admissible velocity for the vehicle maximum velocity [km/h].
        MAX_VEL = 1000

        #: Minimum admissible velocity for the vehicle maximum velocity [km/h].
        MIN_VEL = 1

        #: Delta ratio for calculating the vehicle maximum velocity [km/h].
        DELTA_VEL = 0.1

        #: Full load curve percentage fro calculating the available power [-].
        PREC_FLC = 0.9

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class calculate_maximum_velocity_v2(co2_utl.Constants):
        #: Maximum admissible velocity for the vehicle maximum velocity [km/h].
        MAX_VEL = 1000

        #: Minimum admissible velocity for the vehicle maximum velocity [km/h].
        MIN_VEL = 1

        #: Delta ratio for calculating the vehicle maximum velocity [km/h].
        DELTA_VEL = 1

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class calculate_first_gear_box_ratio(co2_utl.Constants):
        #: Starting slope [-].
        STARTING_SLOPE = 0.5  # --> 50%

        #: Percentage of maximum engine torque to calculate the first gear [-].
        MAX_TORQUE_PERCENTAGE = 0.95

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class design_gear_box_ratios(co2_utl.Constants):
        #: Two factor to design the gear box ratios [-].
        f_two = [
            1, 1.01, 1.02, 1.03, 1.04, 1.05, 1.06, 1.07, 1.08, 1.09, 1.1, 1.11,
            1.12, 1.13, 1.14, 1.15, 1.16, 1.17, 1.18, 1.19, 1.2
        ]

        #: Tuning factor to design the gear box ratios [-].
        f_tuning = [
            1, 1.01, 1.02, 1.03, 1.04, 1.05, 1.06, 1.07, 1.08, 1.09, 1.1
        ]

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class _filter_temperature_samples(co2_utl.Constants):
        #: Max abs val of temperature derivatives during the cold phase [°C/s].
        max_abs_dt_cold = 0.7

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class _rescaling_matrix(co2_utl.Constants):
        #: Percentage width top base (i.e., short base) [-].
        a = 0.9902
        #: Percentage to define the bottom base (i.e., long base) from the phase
        #: corner [-].
        b = 0.1699

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class calculate_aerodynamic_drag_coefficient_v1(co2_utl.Constants):
        #: Aerodynamic drag coefficients function of vehicle body [-].
        cw = {
            'cabriolet': 0.28, 'sedan': 0.27, 'hatchback': 0.3,
            'stationwagon': 0.28, 'suv/crossover': 0.35, 'mpv': 0.3,
            'coupé': 0.27, 'bus': 0.35, 'bestelwagen': 0.35,
            'pick-up': 0.4  # estimated.
        }

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class calculate_aerodynamic_drag_coefficient(co2_utl.Constants):
        #: Aerodynamic drag coefficients function of vehicle category [-].
        cw = {
            'A': 0.34, 'B': 0.31, 'C': 0.29, 'D': 0.30, 'E': 0.30, 'F': 0.28,
            'S': 0.29, 'M': 0.32, 'J': 0.38
        }

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class calculate_f2(co2_utl.Constants):
        #: Deteriorating coefficient of the aerodynamic drag and frontal area
        #: due to the roof box [-].
        roof_box = 1.2

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class calculate_rolling_resistance_coeff(co2_utl.Constants):
        #: Rolling resistance coeff, function of tyre class and category [-].
        #: http://eur-lex.europa.eu/legal-content/EN/TXT/PDF/?uri=CELEX:
        #: 02009R1222-20120530&from=EN
        #: (table A4/1 of eu legislation not world wide)
        coeff = {
            'C1': {
                'A': 5.9, 'B': 7.1, 'C': 8.4, 'D': 9.8, 'E': 9.8, 'F': 11.3,
                'G': 12.9
            },
            'C2': {
                'A': 4.9, 'B': 6.1, 'C': 7.4, 'D': 8.6, 'E': 8.6, 'F': 9.9,
                'G': 11.2
            },
            'C3': {
                'A': 3.5, 'B': 4.5, 'C': 5.5, 'D': 6.5, 'E': 7.5, 'F': 8.5,
                'G': 8.5
            }
        }

        coeff = {k: {i: j / 1000.0 for i, j in v.items()}
                 for k, v in coeff.items()}

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class calculate_f1(co2_utl.Constants):
        #: Linear model coefficients.
        qm = 2.7609 / 2, -71.735 / 2

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class calculate_raw_frontal_area_v1(co2_utl.Constants):
        #: Frontal area formulas function of vehicle_mass [function].
        formulas = dict.fromkeys(
            'ABCDEFS', '0.4041 * np.log(vehicle_mass) - 0.338'
        )
        formulas['J'] = '0.0007 * vehicle_mass + 1.8721'
        formulas['M'] = formulas['J']

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class calculate_frontal_area(co2_utl.Constants):
        #: Projection factor from the row frontal area (h * w) [-].
        projection_factor = 0.84

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class select_prediction_data(co2_utl.Constants):
        #: WLTP calibration data used to predict.
        prediction_data = [
            ['calibration', [
                'angle_slope', 'alternator_nominal_voltage',
                'alternator_efficiency', 'battery_capacity', 'cycle_type',
                'cycle_name', 'engine_capacity', 'engine_stroke',
                'final_drive_efficiency', 'final_drive_ratios', 'frontal_area',
                'final_drive_ratio', 'engine_thermostat_temperature',
                'aerodynamic_drag_coefficient', 'fuel_type', 'ignition_type',
                'gear_box_type', 'engine_max_power',
                'engine_speed_at_max_power', 'rolling_resistance_coeff',
                'time_cold_hot_transition', 'engine_idle_fuel_consumption',
                'engine_type', 'has_start_stop', 'engine_is_turbo',
                'engine_fuel_lower_heating_value', 'f0',
                'has_energy_recuperation', 'fuel_carbon_content_percentage',
                'f1', 'f2', 'full_load_speeds', 'plateau_acceleration',
                'full_load_powers', 'fuel_saving_at_strategy',
                'stand_still_torque_ratio', 'lockup_speed_ratio',
                'change_gear_window_width', 'alternator_start_window_width',
                'stop_velocity', 'min_time_engine_on_after_start',
                'min_engine_on_speed', 'max_velocity_full_load_correction',
                'is_hybrid', 'tyre_code', 'engine_has_cylinder_deactivation',
                'active_cylinder_ratios', 'engine_has_variable_valve_actuation',
                'has_torque_converter', 'has_gear_box_thermal_management',
                'has_lean_burn', 'ki_multiplicative', 'n_wheel_drive',
                'ki_additive', 'has_periodically_regenerating_systems',
                'n_dyno_axes', 'has_selective_catalytic_reduction',
                'start_stop_activation_time', 'has_exhausted_gas_recirculation',
                'engine_n_cylinders', 'initial_drive_battery_state_of_charge',
                'motor_p0_speed_ratio', 'motor_p1_speed_ratio',
                'motor_p2_speed_ratio', 'motor_p2_planetary_speed_ratio',
                'motor_p3_front_speed_ratio', 'motor_p3_rear_speed_ratio',
                'motor_p4_front_speed_ratio', 'motor_p4_rear_speed_ratio',
                'rcb_correction', 'vehicle_mass', 'speed_distance_correction',
                'atct_family_correction_factor', 'is_plugin'
            ]],
            ['models', 'all'],
            ['user', 'all']
        ]

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class _tech_mult_factors(co2_utl.Constants):
        #: Multiplication factors of the engine parameters [-].
        factors = {
            'vva': {
                # 0: {},
                1: {'a': 0.98, 'l': 0.92},
            },
            'lb': {
                # 0: {},
                1: {'a': 1.1, 'b': 0.72, 'c': 0.76, 'a2': 1.25, 'l2': 2.85}
            },
            'egr': {
                # 0: {},
                1: {'a': 1.02, 'b': 1.1, 'c': 1.5, 'a2': 1.1},  # positive turbo
                2: {'a': 1.02, 'b': 1.1, 'c': 1.5, 'a2': 1.1},
                # positive natural aspiration
                3: {'b': 1.08, 'c': 1.15, 'a2': 1.1},  # compression
                4: {'b': 1.08, 'c': 1.15, 'a2': 1.1}  # compression + scr
            }
        }

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class FMEP_egr(co2_utl.Constants):
        #: Exhausted gas recirculation multiplication factors ids [-].
        egr_fact_map = {
            ('positive turbo', False): 1,
            ('positive natural aspiration', False): 2,
            ('compression', False): 3,
            ('compression', True): 4
        }

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class identify_co2_emissions(co2_utl.Constants):
        #: Number of perturbations to identify the co2_emissions [-].
        n_perturbations = 100

        #: Enable first step in the co2_params calibration? [-]
        enable_first_step = True

        #: Enable second step in the co2_params calibration? [-]
        enable_second_step = True

        #: Enable third step co2_params calibration in perturbation loop? [-]
        enable_third_step = False

        #: Absolute error in k_refactor between iterations that is acceptable
        #: for convergence in perturbation loop [-].
        xatol = 1e-4

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class calibrate_co2_params(co2_utl.Constants):
        #: Enable first step in the co2_params calibration? [-]
        enable_first_step = False

        #: Enable second step in the co2_params calibration? [-]
        enable_second_step = False

        #: Enable third step in the co2_params calibration? [-]
        enable_third_step = True

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class BatteryStatusModel(co2_utl.Constants):
        #: Minimum delta time to consider valid a charging state to fit charges
        #: boundaries [s].
        min_delta_time_boundaries = 5

        #: Minimum acceptance percentile to fit the bers threshold [%].
        min_percentile_bers = 90

        #: Minimum delta soc to set the charging boundaries [%].
        min_delta_soc = 8

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class default_ki_multiplicative(co2_utl.Constants):
        #: Multiplicative correction for vehicles with periodically regenerating
        #: systems [-].
        ki_multiplicative = {True: 1.05, False: 1.0}

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class define_fmep_model(co2_utl.Constants):
        #: Percentage of max full bmep curve used as limit in cylinder
        #: deactivation strategy [-].
        acr_full_bmep_curve_percentage = 0.72

        #: Percentage of max mean piston speeds used as upper limit in cylinder
        #: deactivation strategy [-].
        acr_max_mean_piston_speeds_percentage = 0.45

        #: Percentage of min mean piston speeds used as lower limit in cylinder
        #: deactivation strategy [-].
        acr_min_mean_piston_speeds_percentage = 1.5

        #: Percentage of max full bmep curve used as limit in lean burn
        #: strategy [-].
        lb_full_bmep_curve_percentage = 0.4

        #: Percentage of max mean piston speeds used as limit in lean burn
        #: strategy [-].
        lb_max_mean_piston_speeds_percentage = 0.6

        #: Percentage of max full bmep curve used as limit in exhausted gas
        #: recirculation strategy [-].
        egr_full_bmep_curve_percentage = 0.5

        #: Percentage of max mean piston speeds used as limit in exhausted gas
        #: recirculation strategy [-].
        egr_max_mean_piston_speeds_percentage = 0.5

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class define_idle_model_detector(co2_utl.Constants):
        #: eps parameter of DBSCAN [RPM].
        EPS = 100.0

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class identify_idle_engine_speed_std(co2_utl.Constants):
        #: Min standard deviation value [RPM].
        MIN_STD = 100.0

        #: Max standard deviation percentage of median value [-].
        MAX_STD_PERC = 0.3

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class StartStopModel(co2_utl.Constants):
        #: Maximum allowed velocity to stop the engine [km/h].
        stop_velocity = 2.0

        #: Minimum acceleration to switch on the engine [m/s2].
        plateau_acceleration = 1 / 3.6

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class correct_constant_velocity(co2_utl.Constants):
        #: Constant velocities to correct the upper limits for NEDC [km/h].
        CON_VEL_UP_SHIFT = (15.0, 32.0, 50.0, 70.0)

        #: Window to identify if the shifting matrix has limits close to
        # `CON_VEL_UP_SHIFT` [km/h].
        VEL_UP_WINDOW = 3.5

        #: Delta to add to the limit if this is close to `CON_VEL_UP_SHIFT`
        # [km/h].
        DV_UP_SHIFT = -0.5

        #: Constant velocities to correct the bottom limits for NEDC[km/h].
        CON_VEL_DN_SHIFT = (35.0, 50.0)

        #: Window to identify if the shifting matrix has limits close to
        # `CON_VEL_DN_SHIFT` [km/h].
        VEL_DN_WINDOW = 3.5

        #: Delta to add to the limit if this is close to `CON_VEL_DN_SHIFT`
        # [km/h].
        DV_DN_SHIFT = -1

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class define_initial_co2_emission_model_params_guess(co2_utl.Constants):
        #: Initial guess CO2 emission model params.
        CO2_PARAMS = {
            'positive turbo': [
                ('a', {'value': 0.468678, 'min': 0.0}),
                # 'min': 0.398589, 'max': 0.538767},
                ('b', {'value': 0.011859}),  # 'min': 0.006558, 'max': 0.01716},
                ('c', {'value': -0.00069}),
                # 'min': -0.00099, 'max': -0.00038},
                ('a2', {'value': -0.00266, 'max': 0.0}),
                # 'min': -0.00354, 'max': -0.00179},
                ('b2', {'value': 0, 'min': -1, 'max': 1, 'vary': False}),
                ('l', {'value': -2.14063, 'max': 0.0}),
                # 'min': -3.17876, 'max': -1.1025}),
                ('l2', {'value': -0.0025, 'max': 0.0}),
                # 'min': -0.00796, 'max': 0.0}),b
                ('t1', {'value': 3.5, 'min': 0.0, 'max': 8.0}),
                ('dt', {'value': 1.0, 'min': 0.0}),
                ('t0', {'expr': 't1 + dt', 'value': 4.5, 'max': 8.0}),
            ],
            'positive natural aspiration': [
                ('a', {'value': 0.4851, 'min': 0.0}),
                # 'min': 0.40065, 'max': 0.54315},
                ('b', {'value': 0.01193}),  # 'min': -0.00247, 'max': 0.026333},
                ('c', {'value': -0.00065}),
                # 'min': -0.00138, 'max': 0.0000888},
                ('a2', {'value': -0.00385, 'max': 0.0}),
                # 'min': -0.00663, 'max': -0.00107},
                ('b2', {'value': 0, 'min': -1, 'max': 1, 'vary': False}),
                ('l', {'value': -2.39882, 'max': 0.0}),
                # 'min': -3.27698, 'max': -1.72066},
                ('l2', {'value': -0.00286, 'max': 0.0}),
                # 'min': -0.00577, 'max': 0.0},
                ('t1', {'value': 3.5, 'min': 0.0, 'max': 8.0}),
                ('dt', {'value': 1.0, 'min': 0.0}),
                ('t0', {'expr': 't1 + dt', 'value': 4.5, 'max': 8.0}),
            ],
            'compression': [
                ('a', {'value': 0.391197, 'min': 0.0}),
                # 'min': 0.346548, 'max': 0.435846},
                ('b', {'value': 0.028604}),
                # 'min': 0.002519, 'max': 0.054688},
                ('c', {'value': -0.00196}),
                # 'min': -0.00386, 'max': -0.000057},
                ('a2', {'value': -0.0012, 'max': 0.0}),
                # 'min': -0.00233, 'max': -0.000064},
                ('b2', {'value': 0, 'min': -1, 'max': 1, 'vary': False}),
                ('l', {'value': -1.55291, 'max': 0.0}),
                # 'min': -2.2856, 'max': -0.82022},
                ('l2', {'value': -0.0076, 'max': 0.0}),
                # 'min': -0.01852, 'max': 0.0},
                ('t1', {'value': 3.5, 'min': 0.0, 'max': 8.0}),
                ('dt', {'value': 1.0, 'min': 0.0}),
                ('t0', {'expr': 't1 + dt', 'value': 4.5, 'max': 8.0}),
            ]
        }

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class default_specific_gear_shifting(co2_utl.Constants):
        #: Specific gear shifting model.
        SPECIFIC_GEAR_SHIFTING = 'ALL'

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class default_clutch_k_factor_curve(co2_utl.Constants):
        #: Torque ratio when speed ratio==0 for clutch model.
        STAND_STILL_TORQUE_RATIO = 1.0

        #: Minimum speed ratio where torque ratio==1 for clutch model.
        LOCKUP_SPEED_RATIO = 0.0

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class default_tc_k_factor_curve(co2_utl.Constants):
        #: Torque ratio when speed ratio==0 for torque converter model.
        STAND_STILL_TORQUE_RATIO = 1.9

        #: Minimum speed ratio where torque ratio==1 for torque converter model.
        LOCKUP_SPEED_RATIO = 0.87

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class select_default_n_dyno_axes(co2_utl.Constants):
        #: Number of dyno axes [-].
        DYNO_AXES = {
            'WLTP': {2: 1, 4: 2},
            'NEDC': {2: 1, 4: 1}
        }

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class select_phases_integration_times(co2_utl.Constants):
        #: Cycle phases integration times [s].
        INTEGRATION_TIMES = {
            'WLTP': (0.0, 590.0, 1023.0, 1478.0, 1800.0),
            'NEDC': (0.0, 780.0, 1180.0)
        }

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class get_gear_box_efficiency_constants(co2_utl.Constants):
        #: Vehicle gear box efficiency constants (gbp00, gbp10, and gbp01).
        PARAMS = {
            True: {
                'gbp00': {'m': 0.0043233434399999994,
                          'q': {'hot': 0.29823614099999995,
                                'cold': 0.695884329}},
                'gbp10': {'m': 2.4525999999999996e-06,
                          'q': {'hot': 0.0001547871,
                                'cold': 0.0005171569}},
                'gbp01': {'q': {'hot': 0.9793688500000001,
                                'cold': 0.96921995}},
            },
            False: {
                'gbp00': {'m': 0.0043233434399999994,
                          'q': {'hot': 0.29823614099999995,
                                'cold': 0.695884329}},
                'gbp10': {'m': 2.4525999999999996e-06,
                          'q': {'hot': 5.15957e-05,
                                'cold': 0.00012958919999999998}},
                'gbp01': {'q': {'hot': 0.9895177500000001,
                                'cold': 0.9793688500000001}},
            }
        }

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class calculate_engine_mass(co2_utl.Constants):
        #: Equivalent gear box heat capacity parameters.
        PARAMS = {
            'mass_coeff': {
                'compression': 1.1,
                'positive': 1.0
            },
            'mass_reg_coeff': (0.4208, 60)
        }

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class calculate_engine_heat_capacity(co2_utl.Constants):
        #: Equivalent gear box heat capacity parameters.
        PARAMS = {
            'heated_mass_percentage': {
                'coolant': 0.04,  # coolant: 50%/50% (0.85*4.186)
                'oil': 0.055,  # oil: lubricant
                'crankcase': 0.18,  # crankcase: cast iron
                'cyl_head': 0.09,  # cyl_head: aluminium
                'pistons': 0.025,  # pistons: aluminium
                'crankshaft': 0.08,  # crankshaft: steel
                'body': 0.1  # body: cast iron
            },
            # Cp in (J/kgK)
            'heat_capacity': {
                'coolant': 0.85 * 4186.0,
                'oil': 2090.0,
                'crankcase': 460.0,
                'cyl_head': 910.0,
                'pistons': 910.0,
                'crankshaft': 490.0,
                'body': 460.0
            }
        }

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class calculate_equivalent_gear_box_heat_capacity(co2_utl.Constants):
        #: Equivalent gear box heat capacity parameters.
        PARAMS = {
            'gear_box_mass_engine_ratio': 0.25,
            # Cp in (J/kgK)
            'heat_capacity': {
                'oil': 2090.0,
            },
            'thermal_management_factor': 0.5
        }

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class default_full_load_speeds_and_powers(co2_utl.Constants):
        #: Vehicle normalized full load curve.
        FULL_LOAD = {
            'positive': (
                [0., 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1., 1.1, 1.2],
                [0.1, 0.198238659, 0.30313392, 0.410104642, 0.516920841,
                 0.621300767, 0.723313491, 0.820780368, 0.901750158,
                 0.962968496, 0.995867804, 0.953356174, 0.85]
            ),
            'compression': (
                [0., 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1., 1.1, 1.2],
                [0.1, 0.278071182, 0.427366185, 0.572340499, 0.683251935,
                 0.772776746, 0.846217049, 0.906754984, 0.94977083, 0.981937981,
                 1, 0.937598144, 0.85]
            )
        }

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class calculate_engine_max_torque(co2_utl.Constants):
        #: Engine nominal torque params.
        PARAMS = {
            'positive': 1.25,
            'compression': 1.1
        }

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class calculate_engine_moment_inertia(co2_utl.Constants):
        #: Engine moment of inertia params.
        PARAMS = {
            'positive': 1,
            'compression': 2
        }

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class calculate_co2_emissions(co2_utl.Constants):
        # idle ratio to define the fuel cutoff [-].
        cutoff_idle_ratio = 1.1

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class default_initial_drive_battery_state_of_charge(co2_utl.Constants):
        # default initial state of charge of the drive battery [%].
        initial_state_of_charge = {
            'none': sh.NONE,
            'mild': 50,
            'full': 60,
            'plugin': 70,
            'electric': 80
        }

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class default_initial_service_battery_state_of_charge(co2_utl.Constants):
        # default initial state of charge of the service battery [%].
        initial_state_of_charge = {
            'WLTP': 90,
            'NEDC': 99
        }

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class default_fuel_density(co2_utl.Constants):
        #: Fuel density [g/l].
        FUEL_DENSITY = {
            'gasoline': 745.0,
            'diesel': 832.0,
            'LPG': 43200.0 / 46000.0 * 745.0,  # Gasoline equivalent.
            'NG': 43200.0 / 45100.0 * 745.0,  # Gasoline equivalent.
            'ethanol': 794.0,
            'methanol': 791.0,
            'propane': 510.0,
            'biodiesel': 890.0,
        }

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class default_fuel_lower_heating_value(co2_utl.Constants):
        #: Fuel lower heating value [kJ/kg].
        LHV = {
            'gasoline': 43200.0,
            'diesel': 43100.0,
            'LPG': 46000.0,
            'NG': 45100.0,
            'ethanol': 26800.0,
            'methanol': 19800.0,
            'propane': 49680.0,
            'biodiesel': 37900.0,
        }

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class default_fuel_carbon_content(co2_utl.Constants):
        #: Fuel carbon content [CO2g/g].
        CARBON_CONTENT = {
            'gasoline': 3.17,
            'diesel': 3.16,
            'LPG': 1.35,
            'NG': 3.21,
            'ethanol': 1.91,
            'methanol': 1.37,
            'propane': 2.99,
            'biodiesel': 2.81,
        }

    # noinspection PyMissingOrEmptyDocstring,PyPep8Naming
    class _rcb_correction(co2_utl.Constants):
        #: Willans factors [gCO2/MJ].
        WILLANS = {
            'gasoline': {
                'positive turbo': 184, 'positive natural aspiration': 174
            },
            'diesel': {'compression': 161},
            'LPG': {'positive turbo': 164, 'positive natural aspiration': 155},
            'NG': {'positive turbo': 137, 'positive natural aspiration': 129},
            'ethanol': {
                'positive turbo': 179, 'positive natural aspiration': 169
            },
            'methanol': {
                'positive turbo': 184 * 1.37 / 3.17 * 432.0 / 198.0,
                'positive natural aspiration': 174 * 1.37 / 3.17 * 432.0 / 198.0
            },
            'propane': {
                'positive turbo': 184 * 2.99 / 3.17 * 432.0 / 496.8,
                'positive natural aspiration': 174 * 2.99 / 3.17 * 432.0 / 496.8
            },
            'biodiesel': {'compression': 161 * 2.81 / 3.16 * 431.0 / 379.0},
        }


# noinspection PyPep8Naming,PyMissingOrEmptyDocstring
class Defaults(co2_utl.Constants):
    values = Values()
    functions = Functions()

    #: Machine error.
    EPS = 1.1920929e-07

    #: Infinite value.
    INF = 10000.0


dfl = Defaults()



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to produce output report and summary.
"""
import functools
import collections
import numpy as np
import schedula as sh

dsp = sh.BlueDispatcher(
    name='make_report',
    description='Produces a vehicle report from CO2MPAS outputs.'
)


def _split_by_data_format(data):
    d = {}
    p = ('full_load_speeds', 'full_load_torques', 'full_load_powers')
    try:
        s = max(v.size for k, v in data.items()
                if k not in p and isinstance(v, np.ndarray))
    except ValueError:
        s = None

    get_d = functools.partial(
        sh.get_nested_dicts, d, default=collections.OrderedDict
    )

    for k, v in data.items():
        if isinstance(v, np.ndarray) and s == v.size:  # series
            get_d('ts')[k] = v
        else:  # params
            get_d('pa')[k] = v

    return d


def _is_equal(v, iv):
    try:
        if v == iv:
            return True
    except ValueError:
        # noinspection PyUnresolvedReferences
        if (v == iv).all():
            return True
    return False


def _re_sample_targets(data):
    res = {}
    for k, v in sh.stack_nested_keys(data.get('target', {}), depth=2):
        if sh.are_in_nested_dicts(data, 'output', *k):
            o = sh.get_nested_dicts(data, 'output', *k)
            o = _split_by_data_format(o)
            t = sh.selector(o, _split_by_data_format(v), allow_miss=True)

            if 'times' not in t.get('ts', {}) or 'times' not in o['ts']:
                t.pop('ts', None)
            else:
                time_series = t['ts']
                x, xp = o['ts']['times'], time_series.pop('times')
                if not _is_equal(x, xp):
                    for i, fp in time_series.items():
                        time_series[i] = np.interp(x, xp, fp)
            v = sh.combine_dicts(*t.values())
            sh.get_nested_dicts(res, *k[:-1])[k[-1]] = v

    return res


def _correlation_coefficient(t, o):
    with np.errstate(divide='ignore', invalid='ignore'):
        return np.corrcoef(t, o)[0, 1] if t.size > 1 else np.nan


def _prediction_target_ratio(t, o):
    with np.errstate(divide='ignore', invalid='ignore'):
        return np.mean(o / t)


@functools.lru_cache(None)
def _get_metrics():
    from co2mpas.utils import mae
    import sklearn.metrics as sk_met
    metrics = {
        'mean_absolute_error': mae,
        'correlation_coefficient': _correlation_coefficient,
        'accuracy_score': sk_met.accuracy_score,
        'prediction_target_ratio': _prediction_target_ratio
    }
    return metrics


def _compare(t, o, metrics):
    res = {}

    def _asarray(*x):
        x = np.asarray(x)
        if x.dtype is np.dtype(np.bool):
            x = np.asarray(x, dtype=int)
        return x

    try:
        t, o = _asarray(t), _asarray(o)
        for k, v in metrics.items():
            # noinspection PyBroadException
            try:
                m = v(t, o)
                if not np.isnan(m):
                    res[k] = m
            except Exception:
                pass
    except (ValueError, TypeError):
        pass
    return res


def _compare_outputs_vs_targets(data):
    res = {}
    metrics = _get_metrics()
    with np.errstate(divide='ignore', invalid='ignore'):
        for k, t in sh.stack_nested_keys(data.get('target', {}), depth=3):
            if not sh.are_in_nested_dicts(data, 'output', *k):
                continue

            o = sh.get_nested_dicts(data, 'output', *k)
            v = _compare(t, o, metrics=metrics)
            if v:
                sh.get_nested_dicts(res, *k[:-1])[k[-1]] = v

    return res


def _get_values(data, keys, tag=(), update=lambda k, v: v, base=None):
    k = ('input', 'target', 'output')
    data = sh.selector(k, data, allow_miss=True)

    base = {} if base is None else base
    for k, v in sh.stack_nested_keys(data, depth=3):
        k = k[::-1]
        v = sh.selector(keys, v, allow_miss=True)
        v = update(k, v)

        if v:
            sh.get_nested_dicts(base, *tag, *k[:-1])[k[-1]] = v

    return base


def _get_phases_values(data, what='co2_emission', base=None):
    p_wltp, p_nedc = ('low', 'medium', 'high', 'extra_high'), ('UDC', 'EUDC')
    keys = tuple('_'.join((what, v)) for v in (p_wltp + p_nedc + ('value',)))
    keys += ('phases_%ss' % what,)

    def _update(k, v):
        if keys[-1] in v:
            o = v.pop(keys[-1])
            _map = p_nedc if 'nedc' in k[0] else p_wltp
            if len(_map) != len(o):
                v.update(_format_dict(enumerate(o), '{} phase %d'.format(what)))
            else:
                v.update(_format_dict(zip(_map, o), '{}_%s'.format(what)))
        return v

    return _get_values(data, keys, tag=(what,), update=_update, base=base)


def _get_summary_results(data):
    res = {}
    for k in ('declared_co2_emission', 'corrected_co2_emission', 'co2_emission',
              'fuel_consumption', 'declared_sustaining_co2_emission',
              'corrected_sustaining_co2_emission'):
        _get_phases_values(data, what=k, base=res)
    keys = ('f0', 'f1', 'f2', 'vehicle_mass', 'gear_box_type', 'has_start_stop',
            'r_dynamic', 'ki_multiplicative', 'ki_addittive', 'fuel_type',
            'engine_capacity', 'engine_is_turbo', 'engine_max_power',
            'engine_speed_at_max_power', 'drive_battery_delta_state_of_charge',
            'service_battery_delta_state_of_charge', 'is_hybrid')
    _get_values(data, keys, tag=('vehicle',), base=res)

    return res


def _format_selection(score_by_model, depth=-1, index='model_id'):
    res = {}
    for k, v in sorted(sh.stack_nested_keys(score_by_model, depth=depth)):
        v = v.copy()
        v[index] = k[0]
        sh.get_nested_dicts(res, *k[1:], default=list).append(v)
    return res


def _get_selection(data):
    n = ('data', 'calibration', 'model_scores', 'model_selections')
    if sh.are_in_nested_dicts(data, *n):
        return _format_selection(sh.get_nested_dicts(data, *n), 3)
    return {}


def _format_report_summary(data):
    summary = {}
    comparison = _compare_outputs_vs_targets(data)
    if comparison:
        summary['comparison'] = comparison

    selection = _get_selection(data)
    if selection:
        summary['selection'] = selection

    results = _get_summary_results(data)
    if results:
        summary['results'] = results

    return summary


def _add_special_data2report(data, report, to_keys, *from_keys):
    if from_keys[-1] != 'times' and \
            sh.are_in_nested_dicts(data, *from_keys):
        v = sh.get_nested_dicts(data, *from_keys)
        n = to_keys + ('{}.{}'.format(from_keys[0], from_keys[-1]),)
        sh.get_nested_dicts(report, *n[:-1],
                            default=collections.OrderedDict)[n[-1]] = v
        return True, v
    return False, None


def _format_report_output(data):
    res = {}
    func = functools.partial(sh.get_nested_dicts,
                             default=collections.OrderedDict)
    for k, v in sh.stack_nested_keys(data.get('output', {}), depth=3):
        _add_special_data2report(data, res, k[:-1], 'target', *k)

        s, iv = _add_special_data2report(data, res, k[:-1], 'input', *k)
        if not s or (s and not _is_equal(iv, v)):
            func(res, *k[:-1])[k[-1]] = v

    output = {}
    for k, v in sh.stack_nested_keys(res, depth=2):
        v = _split_by_data_format(v)
        sh.get_nested_dicts(output, *k[:-1])[k[-1]] = v

    return output


def _format_scores(scores):
    res = {}
    for k, j in sh.stack_nested_keys(scores, depth=3):
        if k[-1] in ('limits', 'errors'):
            model_id = k[0]
            extra_field = ('score',) if k[-1] == 'errors' else ()
            for i, v in sh.stack_nested_keys(j):
                i = (model_id, i[-1], k[1],) + i[:-1] + extra_field
                sh.get_nested_dicts(res, *i[:-1])[i[-1]] = v
    sco = {}
    for k, v in sorted(sh.stack_nested_keys(res, depth=4)):
        v.update(sh.map_list(['model_id', 'param_id'], *k[:2]))
        sh.get_nested_dicts(sco, *k[2:], default=list).append(v)
    return sco


def _format_report_scores(data):
    res = {}
    scores = 'data', 'calibration', 'model_scores'
    if sh.are_in_nested_dicts(data, *scores):
        n = scores + ('param_selections',)
        v = _format_selection(sh.get_nested_dicts(data, *n), 2, 'param_id')
        if v:
            sh.get_nested_dicts(res, *n[:-1])[n[-1]] = v

        n = scores + ('model_selections',)
        v = _format_selection(sh.get_nested_dicts(data, *n), 3)
        if v:
            sh.get_nested_dicts(res, *n[:-1])[n[-1]] = v

        n = scores + ('score_by_model',)
        v = _format_selection(sh.get_nested_dicts(data, *n), 2)
        if v:
            sh.get_nested_dicts(res, *n[:-1])[n[-1]] = v

        n = scores + ('scores',)
        v = _format_scores(sh.get_nested_dicts(data, *n))
        if v:
            sh.get_nested_dicts(res, *n[:-1])[n[-1]] = v

    return res


def _map_cycle_report_graphs():
    _map = collections.OrderedDict()

    _map['fuel_consumptions'] = {
        'label': 'fuel consumption',
        'set': {
            'title': {'name': 'Fuel consumption'},
            'y_axis': {'name': 'Fuel consumption [g/s]'},
            'x_axis': {'name': 'Time [s]'},
            'legend': {'position': 'bottom'}
        }
    }

    _map['engine_speeds_out'] = {
        'label': 'engine speed',
        'set': {
            'title': {'name': 'Engine speed [RPM]'},
            'y_axis': {'name': 'Engine speed [RPM]'},
            'x_axis': {'name': 'Time [s]'},
            'legend': {'position': 'bottom'}
        }
    }

    _map['engine_powers_out'] = {
        'label': 'engine power',
        'set': {
            'title': {'name': 'Engine power [kW]'},
            'y_axis': {'name': 'Engine power [kW]'},
            'x_axis': {'name': 'Time [s]'},
            'legend': {'position': 'bottom'}
        }
    }

    _map['velocities'] = {
        'label': 'velocity',
        'set': {
            'title': {'name': 'Velocity [km/h]'},
            'y_axis': {'name': 'Velocity [km/h]'},
            'x_axis': {'name': 'Time [s]'},
            'legend': {'position': 'bottom'}
        }
    }

    _map['engine_coolant_temperatures'] = {
        'label': 'engine coolant temperature',
        'set': {
            'title': {'name': 'Engine temperature [°C]'},
            'y_axis': {'name': 'Engine temperature [°C]'},
            'x_axis': {'name': 'Time [s]'},
            'legend': {'position': 'bottom'}
        }
    }

    _map['service_battery_state_of_charges'] = {
        'label': 'SOC',
        'set': {
            'title': {'name': 'Service battery state of charge [%]'},
            'y_axis': {'name': 'Service battery state of charge [%]'},
            'x_axis': {'name': 'Time [s]'},
            'legend': {'position': 'bottom'}
        }
    }

    _map['service_battery_electric_powers'] = {
        'label': 'service battery electric power',
        'set': {
            'title': {'name': 'Service battery electric power [kW]'},
            'y_axis': {'name': 'Service battery electric power [kW]'},
            'x_axis': {'name': 'Time [s]'},
            'legend': {'position': 'bottom'}
        }
    }

    _map['drive_battery_state_of_charges'] = {
        'label': 'SOC',
        'set': {
            'title': {'name': 'Drive battery state of charge [%]'},
            'y_axis': {'name': 'Drive battery state of charge [%]'},
            'x_axis': {'name': 'Time [s]'},
            'legend': {'position': 'bottom'}
        }
    }

    _map['drive_battery_electric_powers'] = {
        'label': 'drive electric power',
        'set': {
            'title': {'name': 'Drive battery power [kW]'},
            'y_axis': {'name': 'Drive battery power [kW]'},
            'x_axis': {'name': 'Time [s]'},
            'legend': {'position': 'bottom'}
        }
    }

    _map['alternator_currents'] = {
        'label': 'alternator current',
        'set': {
            'title': {'name': 'Alternator current [A]'},
            'y_axis': {'name': 'Alternator current [A]'},
            'x_axis': {'name': 'Time [s]'},
            'legend': {'position': 'bottom'}
        }
    }

    _map['gear_box_temperatures'] = {
        'label': 'gear box temperature',
        'set': {
            'title': {'name': 'Gear box temperature [°C]'},
            'y_axis': {'name': 'Gear box temperature [°C]'},
            'x_axis': {'name': 'Time [s]'},
            'legend': {'position': 'bottom'}
        }
    }

    return _map


# noinspection PyProtectedMember
def _get_chart_reference(report):
    from .write.excel import _sheet_name
    from .load.excel import _re_params_name
    r, _map = {}, _map_cycle_report_graphs()
    out = report.get('output', {})
    it = sh.stack_nested_keys(out, key=('output',), depth=3)
    for k, v in sorted(it):
        if k[-1] == 'ts' and 'times' in v:
            label = '{}/%s'.format(_sheet_name(k))
            for i, j in sorted(v.items()):
                param_id = _re_params_name.match(i)['param']
                m = _map.get(param_id, None)
                if m:
                    d = {
                        'x': k + ('times',),
                        'y': k + (i,),
                        'label': label % i
                    }
                    n = k[2], param_id, 'series'
                    sh.get_nested_dicts(r, *n, default=list).append(d)

    for k, v in sh.stack_nested_keys(r, depth=2):
        m = _map[k[1]]
        m.pop('label', None)
        v.update(m)

    return r


@sh.add_function(dsp, outputs=['report'])
def format_report_output_data(output_data):
    """
    Produces a vehicle output report from CO2MPAS outputs.

    :param output_data:
        CO2MPAS outputs.
    :type output_data: dict

    :return:
        Vehicle output report.
    :rtype: dict
    """
    output_data = output_data.copy()

    report = {}

    if 'pipe' in output_data:
        report['pipe'] = output_data['pipe']

    target = _re_sample_targets(output_data)
    if target:
        output_data['target'] = target

    summary = _format_report_summary(output_data)
    if summary:
        report['summary'] = summary

    output = _format_report_output(output_data)
    if output:
        report['output'] = output

    scores = _format_report_scores(output_data)
    if scores:
        sh.combine_nested_dicts(scores, base=report)

    graphs = _get_chart_reference(report)
    if graphs:
        report['graphs'] = graphs

    return report


def _extract_summary_from_summary(report, extracted, augmented_summary=False):
    n = ('summary', 'results')
    keys = (
        'corrected_sustaining_co2_emission', 'declared_sustaining_co2_emission',
        'declared_co2_emission', 'co2_emission', 'corrected_co2_emission'
    )
    if augmented_summary:
        keys += 'fuel_consumption', 'vehicle'
    if sh.are_in_nested_dicts(report, *n):
        for j, w in sh.get_nested_dicts(report, *n).items():
            if j in keys:
                for k, v in sh.stack_nested_keys(w, depth=3):
                    if v:
                        sh.get_nested_dicts(extracted, *k).update(v)
    n = ('summary', 'comparison')
    if augmented_summary and sh.are_in_nested_dicts(report, *n):
        comp = sh.get_nested_dicts(report, *n)
        for k, v in sh.stack_nested_keys(comp, key=('comparison',), depth=2):
            sh.get_nested_dicts(extracted, *k[::-1]).update({
                '/'.join(i): j for i, j in sh.stack_nested_keys(v)
            })


def _param_names_values(data):
    # noinspection PyProtectedMember
    from .load.excel import _re_params_name
    for k, v in data.items():
        m = _re_params_name.match(k)
        yield m['usage'] or 'output', m['param'], v


def _format_dict(gen, str_format='%s', func=lambda x: x):
    return {str_format % k: func(v) for k, v in gen}


def _extract_summary_from_output(report, extracted, augmented_summary=False):
    for k, v in sh.stack_nested_keys(report.get('output', {}), depth=2):
        k = k[::-1]
        for u, i, j in _param_names_values(v.get('pa', {})):
            o = {}
            if i in ('has_sufficient_power',):
                o = {i: j}
            elif augmented_summary:
                if i == 'co2_params_calibrated':
                    o = _format_dict(j.valuesdict().items(), 'co2_params %s')
                elif i == 'calibration_status':
                    o = _format_dict(
                        enumerate(j), 'status co2_params step %d',
                        lambda x: x[0]
                    )
                elif i == 'willans_factors':
                    o = j
                elif i == 'phases_willans_factors':
                    for n, m in enumerate(j):
                        o.update(_format_dict(
                            m.items(), '%s phase {}'.format(n)
                        ))
                elif i == 'co2_rescaling_scores':
                    o = sh.map_list(
                        ['rescaling_mean', 'rescaling_std', 'rescaling_n'], *j
                    )

            if o:
                sh.get_nested_dicts(extracted, *(k + (u,))).update(o)


def _extract_summary_from_model_scores(
        report, extracted, augmented_summary=False):
    n = ('data', 'calibration', 'model_scores', 'model_selections')
    if not augmented_summary or not sh.are_in_nested_dicts(report, *n):
        return False

    sel = sh.get_nested_dicts(report, *n)
    s = ('data', 'calibration', 'model_scores', 'score_by_model')
    score = sh.get_nested_dicts(report, *s)
    s = ('data', 'calibration', 'model_scores', 'scores')
    scores = sh.get_nested_dicts(report, *s)

    for k, v in sh.stack_nested_keys(extracted, depth=3):
        n = k[1::-1]
        if k[-1] == 'output' and sh.are_in_nested_dicts(sel, *n):
            gen = sh.get_nested_dicts(sel, *n)
            gen = ((d['model_id'], d['status']) for d in gen if 'status' in d)
            o = _format_dict(gen, 'status %s')
            v.update(o)
            if k[1] == 'calibration' and k[0] in score:
                gen = score[k[0]]
                gen = ((d['model_id'], d['score']) for d in gen if 'score' in d)
                o = _format_dict(gen, 'score %s')
                v.update(o)
                for i, j in scores[k[0]].items():
                    gen = (
                        ('/'.join((d['model_id'], d['param_id'])), d['score'])
                        for d in j if 'score' in d
                    )
                    o = _format_dict(gen, 'score {}/%s'.format(i))
                    v.update(o)

    return True


@sh.add_function(dsp, outputs=['summary'], inputs_kwargs=True)
def extract_summary(report, augmented_summary=False):
    """
    Extract a summary report.

    :param report:
        Vehicle output report.
    :type report: dict

    :param augmented_summary:
        Add more outputs to the summary.
    :type augmented_summary: bool

    :return:
        Summary report.
    :rtype: dict
    """
    extracted = {}

    _extract_summary_from_summary(report, extracted, augmented_summary)

    _extract_summary_from_output(report, extracted, augmented_summary)

    _extract_summary_from_model_scores(report, extracted, augmented_summary)

    return extracted
